"""Overlay function that avoids a GEOSException from geopandas.overlay.

This module includes the function 'clean_overlay', which bypasses a
GEOSException from the regular geopandas.overlay. The function is a generalized
version of the solution from GH 2792.

'clean_overlay' also includes the overlay type "update", which can be specified in the
"how" parameter, in addition to the five native geopandas how-s.
"""

import functools
from collections.abc import Callable

import dask.array as da
import geopandas as gpd
import joblib
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from pandas import DataFrame
from shapely import Geometry
from shapely import STRtree
from shapely import box
from shapely import difference
from shapely import intersection
from shapely import make_valid
from shapely import unary_union
from shapely.errors import GEOSException

from .general import _determine_geom_type_args
from .general import clean_geoms
from .geometry_types import get_geom_type
from .geometry_types import make_all_singlepart
from .geometry_types import to_single_geom_type

DEFAULT_GRID_SIZE = None
DEFAULT_LSUFFIX = "_1"
DEFAULT_RSUFFIX = "_2"


def clean_overlay(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    how: str = "intersection",
    keep_geom_type: bool | None = None,
    geom_type: str | None = None,
    predicate: str | None = "intersects",
    grid_size: float | None = None,
    n_jobs: int = 1,
    lsuffix: str = DEFAULT_LSUFFIX,
    rsuffix: str = DEFAULT_RSUFFIX,
) -> GeoDataFrame:
    """Fixes and explodes geometries before doing a shapely overlay, then cleans up.

    Fixes geometries, then does a shapely overlay operation that dodges a GEOSException
    raised in the regular geopandas.overlay. The function is a generalised version of a
    solution from GH 2792.

    Args:
        df1: GeoDataFrame
        df2: GeoDataFrame
        how: Method of spatial overlay. Includes the 'update' method, plus the five
            native geopandas methods 'intersection', 'union', 'identity',
            'symmetric_difference' and 'difference'.
        keep_geom_type: If True (default), return only geometries of the same
            geometry type as df1 has, if False, return all resulting geometries.
        geom_type: optionally specify what geometry type to keep before the overlay,
            if there are mixed geometry types. Must be either "polygon", "line" or
            "point".
        grid_size: Precision grid size to round the geometries. Will use the highest
            precision of the inputs by default.
        n_jobs: number of threads.
        predicate: Spatial predicate in the spatial tree.
        lsuffix: Suffix of columns in df1 that are also in df2.
        rsuffix: Suffix of columns in df2 that are also in df1.

    Returns:
        GeoDataFrame with overlayed and fixed geometries and columns from both
        GeoDataFrames.

    Raises:
        ValueError: If 'how' is not one of 'intersection', 'union', 'identity',
            'symmetric_difference', 'difference' or 'update'.
    """
    # Allowed operations
    allowed_hows = [
        "intersection",
        "union",
        "identity",
        "symmetric_difference",
        "difference",
        "update",
    ]
    if how not in allowed_hows:
        raise ValueError(
            f"`how` was {how!r} but is expected to be in {', '.join(allowed_hows)}"
        )

    if df1.crs != df2.crs:
        raise ValueError(f"'crs' mismatch. Got {df1.crs} and {df2.crs}")

    crs = df1.crs

    # original_geom_type = geom_type

    df1, geom_type, keep_geom_type = _determine_geom_type_args(
        df1, geom_type, keep_geom_type
    )

    if not geom_type or geom_type == "mixed":
        if keep_geom_type and geom_type == "mixed":
            raise ValueError(
                "mixed geometries are not allowed when geom_type isn't specified.",
                df1.geometry.geom_type.value_counts(),
            )

    df1 = clean_geoms(df1)
    df2 = clean_geoms(df2)

    df1 = make_all_singlepart(df1, ignore_index=True)
    df2 = make_all_singlepart(df2, ignore_index=True)

    if keep_geom_type:
        df1 = to_single_geom_type(df1, geom_type)

    if geom_type and get_geom_type(df1) == get_geom_type(df2):
        df2 = to_single_geom_type(df2, geom_type)

    assert df1.is_valid.all(), df1.is_valid.value_counts()
    assert df2.is_valid.all(), df2.is_valid.value_counts()
    assert df1.geometry.notna().all()
    assert df2.geometry.notna().all()

    box1 = box(*df1.total_bounds)
    box2 = box(*df2.total_bounds)

    if not len(df1) or not len(df1) or not box1.intersects(box2):
        return _no_intersections_return(df1, df2, how, lsuffix, rsuffix)

    if df1._geometry_column_name != "geometry":
        df1 = df1.rename_geometry("geometry")

    if df2._geometry_column_name != "geometry":
        df2 = df2.rename_geometry("geometry")

    # to pandas because GeoDataFrame constructor is expensive
    df1 = DataFrame(df1).reset_index(drop=True)
    df2 = DataFrame(df2).reset_index(drop=True)

    overlayed = (
        gpd.GeoDataFrame(
            _shapely_pd_overlay(
                df1,
                df2,
                how=how,
                grid_size=grid_size,
                lsuffix=lsuffix,
                rsuffix=rsuffix,
                geom_type=geom_type,
                n_jobs=n_jobs,
                predicate=predicate,
            ),
            geometry="geometry",
            crs=crs,
        )
        .pipe(clean_geoms)
        .pipe(make_all_singlepart, ignore_index=True)
    )

    if keep_geom_type:
        overlayed = to_single_geom_type(overlayed, geom_type)

    return overlayed.reset_index(drop=True)


def _join_and_get_no_rows(df1, df2, lsuffix, rsuffix):
    geom_col = df1._geometry_column_name
    df1_cols = df1.columns.difference({geom_col})
    df2_cols = df2.columns.difference({df2._geometry_column_name})
    cols_with_suffix = [
        f"{col}{lsuffix}" if col in df2_cols else col for col in df1_cols
    ] + [f"{col}{rsuffix}" if col in df1_cols else col for col in df2_cols]

    return GeoDataFrame(
        pd.DataFrame(columns=cols_with_suffix + [geom_col]),
        geometry=geom_col,
        crs=df1.crs,
    )


def _no_intersections_return(
    df1: GeoDataFrame, df2: GeoDataFrame, how: str, lsuffix, rsuffix: str
) -> GeoDataFrame:
    """Return with no overlay if no intersecting bounding box."""
    if how == "intersection":
        return _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)

    if how == "difference":
        return df1.reset_index(drop=True)

    if how == "identity":
        # add suffixes and return df1
        df_template = _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)
        df2_cols = df2.columns.difference({df2._geometry_column_name})
        df1.columns = [f"{col}{lsuffix}" if col in df2_cols else col for col in df1]
        return pd.concat([df_template, df1], ignore_index=True)

    if how == "update":
        return pd.concat([df1, df2], ignore_index=True)

    assert how in ["union", "symmetric_difference"]

    # add suffixes and return both concatted
    df_template = _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)
    if not len(df1) and not len(df2):
        return df_template

    df_template = _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)
    return pd.concat([df_template, df1, df2], ignore_index=True)


def _shapely_pd_overlay(
    df1: DataFrame,
    df2: DataFrame,
    how: str,
    grid_size: float = DEFAULT_GRID_SIZE,
    predicate: str = "intersects",
    lsuffix: str = DEFAULT_LSUFFIX,
    rsuffix: str = DEFAULT_RSUFFIX,
    geom_type: str | None = None,
    n_jobs: int = 1,
) -> DataFrame:
    if not grid_size and not len(df1) or not len(df2):
        return _no_intersections_return(df1, df2, how, lsuffix, rsuffix)

    tree = STRtree(df2.geometry.values)
    left, right = tree.query(df1.geometry.values, predicate=predicate)

    pairs = _get_intersects_pairs(df1, df2, left, right, rsuffix)
    assert pairs.geometry.notna().all()
    assert pairs.geom_right.notna().all()

    if how == "intersection":
        overlayed = [
            _intersection(
                pairs, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
            )
        ]

    elif how == "difference":
        overlayed = _difference(
            pairs, df1, left, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
        )

    elif how == "symmetric_difference":
        overlayed = _symmetric_difference(
            pairs,
            df1,
            df2,
            left,
            right,
            grid_size=grid_size,
            rsuffix=rsuffix,
            geom_type=geom_type,
            n_jobs=n_jobs,
        )

    elif how == "identity":
        overlayed = _identity(
            pairs, df1, left, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
        )

    elif how == "union":
        overlayed = _union(
            pairs,
            df1,
            df2,
            left,
            right,
            grid_size=grid_size,
            rsuffix=rsuffix,
            geom_type=geom_type,
            n_jobs=n_jobs,
        )

    elif how == "update":
        overlayed = _update(
            pairs,
            df1,
            df2,
            left=left,
            grid_size=grid_size,
            n_jobs=n_jobs,
            geom_type=geom_type,
        )

    assert isinstance(overlayed, list)

    overlayed = pd.concat(overlayed, ignore_index=True).drop(
        columns="_overlay_index_right", errors="ignore"
    )

    # push geometry column to the end
    overlayed = overlayed.reindex(
        columns=[c for c in overlayed.columns if c != "geometry"] + ["geometry"]
    )

    if how not in ["difference", "update"]:
        overlayed = _add_suffix_left(overlayed, df1, df2, lsuffix)

    overlayed["geometry"] = make_valid(overlayed["geometry"])
    # None and empty are falsy
    overlayed = overlayed.loc[lambda x: x["geometry"].notna()]

    return overlayed


def _update(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    left: np.ndarray,
    grid_size: float | None | int,
    geom_type: str | None,
    n_jobs: int,
) -> GeoDataFrame:
    overlayed = _difference(
        pairs, df1, left, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
    )

    return overlayed + [df2]


def _run_overlay_dask(
    arr1: np.ndarray,
    arr2: np.ndarray,
    func: Callable,
    n_jobs: int,
    grid_size: float | int | None,
) -> np.ndarray:
    if len(arr1) // n_jobs <= 1:
        try:
            return func(arr1, arr2, grid_size=grid_size)
        except TypeError as e:
            raise TypeError(e, {type(x) for x in arr1}, {type(x) for x in arr2}) from e
    arr1 = da.from_array(arr1, chunks=len(arr1) // n_jobs)
    arr2 = da.from_array(arr2, chunks=len(arr2) // n_jobs)
    res = arr1.map_blocks(func, arr2, grid_size=grid_size, dtype=float)
    return res.compute(scheduler="threads", optimize_graph=False, num_workers=n_jobs)


def _run_overlay_joblib_threading(
    arr1: np.ndarray,
    arr2: np.ndarray,
    func: Callable,
    n_jobs: int,
    grid_size: int | float | None,
) -> list[Geometry]:
    if len(arr1) // n_jobs <= 1:
        try:
            return func(arr1, arr2, grid_size=grid_size)
        except TypeError as e:
            raise TypeError(e, {type(x) for x in arr1}, {type(x) for x in arr2}) from e
    with joblib.Parallel(n_jobs=n_jobs, backend="threading") as parallel:
        return parallel(
            joblib.delayed(func)(g1, g2, grid_size=grid_size)
            for g1, g2 in zip(arr1, arr2, strict=True)
        )


def _intersection(
    pairs: pd.DataFrame,
    grid_size: None | float | int,
    geom_type: str | None,
    n_jobs: int = 1,
) -> GeoDataFrame:
    if not len(pairs):
        return pairs.drop(columns="geom_right")

    intersections = pairs.copy()

    arr1 = intersections["geometry"].to_numpy()
    arr2 = intersections["geom_right"].to_numpy()

    if n_jobs > 1 and len(arr1) / n_jobs > 10:
        try:
            res = _run_overlay_joblib_threading(
                arr1,
                arr2,
                func=intersection,
                n_jobs=n_jobs,
                grid_size=grid_size,
            )
        except GEOSException:
            arr1 = make_valid_and_keep_geom_type(
                arr1, geom_type=geom_type, n_jobs=n_jobs
            )
            arr2 = make_valid_and_keep_geom_type(
                arr2, geom_type=geom_type, n_jobs=n_jobs
            )
            arr1 = arr1.loc[lambda x: x.index.isin(arr2.index)]
            arr2 = arr2.loc[lambda x: x.index.isin(arr1.index)]

            res = _run_overlay_joblib_threading(
                arr1.to_numpy(),
                arr2.to_numpy(),
                func=intersection,
                n_jobs=n_jobs,
                grid_size=grid_size,
            )
        intersections["geometry"] = res
        return intersections.drop(columns="geom_right")

    try:
        intersections["geometry"] = intersection(
            intersections["geometry"].to_numpy(),
            intersections["geom_right"].to_numpy(),
            grid_size=grid_size,
        )
    except GEOSException:
        left = make_valid_and_keep_geom_type(
            intersections["geometry"].to_numpy(), geom_type, n_jobs=n_jobs
        )
        right = make_valid_and_keep_geom_type(
            intersections["geom_right"].to_numpy(), geom_type, n_jobs=n_jobs
        )
        left = left.loc[lambda x: x.index.isin(right.index)]
        right = right.loc[lambda x: x.index.isin(left.index)]

        intersections["geometry"] = intersection(
            left.to_numpy(), right.to_numpy(), grid_size=grid_size
        )

    return intersections.drop(columns="geom_right")


def _union(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    left: np.ndarray,
    right: np.ndarray,
    grid_size: int | float | None,
    rsuffix: str,
    geom_type: str | None,
    n_jobs: int = 1,
) -> list[GeoDataFrame]:
    merged = []
    if len(left):
        intersections = _intersection(
            pairs, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
        )
        merged.append(intersections)
    symmdiff = _symmetric_difference(
        pairs,
        df1,
        df2,
        left,
        right,
        grid_size=grid_size,
        rsuffix=rsuffix,
        geom_type=geom_type,
        n_jobs=n_jobs,
    )
    merged += symmdiff
    return merged


def _identity(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    left: np.ndarray,
    grid_size: int | float | None,
    geom_type: str | None,
    n_jobs: int = 1,
) -> list[GeoDataFrame]:
    merged = []
    if len(left):
        intersections = _intersection(
            pairs, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
        )
        merged.append(intersections)
    diff = _difference(pairs, df1, left, grid_size=grid_size, n_jobs=n_jobs)
    merged += diff
    return merged


def _symmetric_difference(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    left: np.ndarray,
    right: np.ndarray,
    grid_size: int | float | None,
    rsuffix: str,
    geom_type: str | None,
    n_jobs: int = 1,
) -> list[GeoDataFrame]:
    merged = []

    difference_left = _difference(
        pairs, df1, left, grid_size=grid_size, geom_type=geom_type, n_jobs=n_jobs
    )
    merged += difference_left

    if len(left):
        clip_right = _shapely_diffclip_right(
            pairs,
            df1,
            df2,
            grid_size=grid_size,
            rsuffix=rsuffix,
            geom_type=geom_type,
            n_jobs=n_jobs,
        )
        merged.append(clip_right)

    diff_right = _add_from_right(df1, df2, right, rsuffix)
    merged.append(diff_right)

    return merged


def _difference(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    left: np.ndarray,
    grid_size: int | float | None = None,
    geom_type: str | None = None,
    n_jobs: int = 1,
) -> list[GeoDataFrame]:
    merged = []
    if len(left):
        clip_left = _shapely_diffclip_left(
            pairs=pairs,
            df1=df1,
            grid_size=grid_size,
            geom_type=geom_type,
            n_jobs=n_jobs,
        )
        merged.append(clip_left)
    diff_left = _add_indices_from_left(df1, left)
    merged.append(diff_left)
    return merged


def _get_intersects_pairs(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    left: np.ndarray,
    right: np.ndarray,
    rsuffix: str,
) -> DataFrame:
    return pd.concat(
        [
            df1.take(left),
            (
                DataFrame(
                    {"_overlay_index_right": right}, index=df1.index.values.take(left)
                )
            ),
        ],
        axis=1,
    ).join(
        df2.rename(columns={"geometry": "geom_right"}, errors="raise"),
        on="_overlay_index_right",
        rsuffix=rsuffix,
    )


def _add_suffix_left(
    overlayed: pd.DataFrame, df1: pd.DataFrame, df2: pd.DataFrame, lsuffix: str
):
    """Separating this from _add_indices_from_left, since this suffix is not needed in difference."""
    return overlayed.rename(
        columns={
            c: (
                f"{c}{lsuffix}"
                if c in df1.columns and c in df2.columns and c != "geometry"
                else c
            )
            for c in overlayed.columns
        }
    )


def _add_indices_from_left(df1: pd.DataFrame, left: np.ndarray) -> pd.DataFrame:
    return df1.take(np.setdiff1d(np.arange(len(df1)), left))


def _add_from_right(
    df1: GeoDataFrame, df2: GeoDataFrame, right: np.ndarray, rsuffix: str
) -> GeoDataFrame:
    return df2.take(np.setdiff1d(np.arange(len(df2)), right)).rename(
        columns={
            c: f"{c}{rsuffix}" if c in df1.columns and c != "geometry" else c
            for c in df2.columns
        }
    )


def _shapely_diffclip_left(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    grid_size: int | float | None,
    geom_type: str | None,
    n_jobs: int,
) -> pd.DataFrame:
    """Aggregate areas in right by unique values from left, then erases those from left."""
    keep_cols = list(df1.columns.difference({"_overlay_index_right"})) + ["geom_right"]

    agg_geoms_partial = functools.partial(_agg_geoms, grid_size=grid_size)

    try:
        only_one = pairs.groupby(level=0).transform("size") == 1
        one_hit = pairs.loc[only_one, list(keep_cols)]
        many_hits = pairs.loc[~only_one, list(keep_cols) + ["_overlay_index_right"]]
        # keep first in non-geom-cols, agg only geom col bacause of speed
        many_hits_agged = many_hits.loc[
            lambda x: ~x.index.duplicated(),
            lambda x: x.columns.difference({"geom_right"}),
        ]

        index_mapper = {
            i: x
            for i, x in many_hits.groupby(level=0)["_overlay_index_right"]
            .unique()
            .apply(lambda j: tuple(sorted(j)))
            .items()
        }

        many_hits_agged["_right_indices"] = index_mapper

        inverse_index_mapper = pd.Series(
            {
                x[0]: x
                for x in many_hits_agged.reset_index()
                .groupby("_right_indices")["index"]
                .unique()
                .apply(tuple)
            }
        ).explode()
        inverse_index_mapper = pd.Series(
            inverse_index_mapper.index, index=inverse_index_mapper.values
        )

        agger = (
            pd.Series(index_mapper.values(), index=index_mapper.keys())
            .drop_duplicates()
            .explode()
            .to_frame("_overlay_index_right")
        )
        agger["geom_right"] = agger["_overlay_index_right"].map(
            {
                i: g
                for i, g in zip(
                    many_hits["_overlay_index_right"],
                    many_hits["geom_right"],
                    strict=False,
                )
            }
        )

        agged = pd.Series(
            {
                i: agg_geoms_partial(geoms)
                for i, geoms in agger.groupby(level=0)["geom_right"]
            }
        )
        many_hits_agged["geom_right"] = inverse_index_mapper.map(agged)
        many_hits_agged = many_hits_agged.drop(columns=["_right_indices"])

        clip_left = pd.concat([one_hit, many_hits_agged])
    except IndexError:
        clip_left = pairs.loc[:, list(keep_cols)]

    assert clip_left["geometry"].notna().all()
    assert clip_left["geom_right"].notna().all()

    clip_left["geometry"] = _try_difference(
        clip_left["geometry"].to_numpy(),
        clip_left["geom_right"].to_numpy(),
        grid_size=grid_size,
        geom_type=geom_type,
        n_jobs=n_jobs,
    )

    return clip_left.drop(columns="geom_right")


def _shapely_diffclip_right(
    pairs: pd.DataFrame,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    grid_size: int | float | None,
    rsuffix: str,
    geom_type: str | None,
    n_jobs: int,
) -> pd.DataFrame:
    agg_geoms_partial = functools.partial(_agg_geoms, grid_size=grid_size)

    pairs = pairs.rename(columns={"geometry": "geom_left", "geom_right": "geometry"})

    try:
        only_one = pairs.groupby("_overlay_index_right").transform("size") == 1
        one_hit = pairs[only_one].set_index("_overlay_index_right")[
            ["geom_left", "geometry"]
        ]
        many_hits = (
            pairs[~only_one]
            .groupby("_overlay_index_right")
            .agg(
                {
                    "geom_left": agg_geoms_partial,
                    "geometry": "first",
                }
            )
        )
        clip_right = (
            pd.concat([one_hit, many_hits])
            .join(df2.drop(columns=["geometry"]))
            .rename(
                columns={
                    c: f"{c}{rsuffix}" if c in df1.columns and c != "geometry" else c
                    for c in df2.columns
                }
            )
        )
    except IndexError:
        clip_right = pairs.join(df2.drop(columns=["geometry"])).rename(
            columns={
                c: f"{c}{rsuffix}" if c in df1.columns and c != "geometry" else c
                for c in df2.columns
            }
        )

    assert clip_right["geometry"].notna().all()
    assert clip_right["geom_left"].notna().all()

    clip_right["geometry"] = _try_difference(
        clip_right["geometry"].to_numpy(),
        clip_right["geom_left"].to_numpy(),
        grid_size=grid_size,
        geom_type=geom_type,
    )

    return clip_right.drop(columns="geom_left")


def _try_difference(
    left: np.ndarray,
    right: np.ndarray,
    grid_size: int | float | None,
    geom_type: str | None,
    n_jobs: int = 1,
) -> np.ndarray:
    """Try difference overlay, then make_valid and retry."""
    if n_jobs > 1 and len(left) / n_jobs > 10:
        try:
            return _run_overlay_joblib_threading(
                left,
                right,
                func=difference,
                n_jobs=n_jobs,
                grid_size=grid_size,
            )
        except GEOSException:
            left = make_valid_and_keep_geom_type(
                left, geom_type=geom_type, n_jobs=n_jobs
            )
            right = make_valid_and_keep_geom_type(
                right, geom_type=geom_type, n_jobs=n_jobs
            )
            left = left.loc[lambda x: x.index.isin(right.index)]
            right = right.loc[lambda x: x.index.isin(left.index)]

            return _run_overlay_joblib_threading(
                left.to_numpy(),
                right.to_numpy(),
                func=difference,
                n_jobs=n_jobs,
                grid_size=grid_size,
            )

    try:
        return difference(
            left,
            right,
            grid_size=grid_size,
        )
    except GEOSException:
        left = make_valid_and_keep_geom_type(left, geom_type, n_jobs=n_jobs)
        right = make_valid_and_keep_geom_type(right, geom_type, n_jobs=n_jobs)
        left = left.loc[lambda x: x.index.isin(right.index)]
        right = right.loc[lambda x: x.index.isin(left.index)]
        try:
            return difference(
                left.to_numpy(),
                right.to_numpy(),
                grid_size=grid_size,
            )
        except GEOSException as e:
            raise e.__class__(e, f"{grid_size=}", f"{left=}", f"{right=}") from e


def make_valid_and_keep_geom_type(
    geoms: np.ndarray, geom_type: str, n_jobs: int
) -> GeoSeries:
    """Make GeometryCollections into (Multi)Polygons, (Multi)LineStrings or (Multi)Points.

    Because GeometryCollections might appear after dissolving (unary_union).
    And this makes shapely difference/intersection fail.

    Args:
        geoms: Array of geometries.
        geom_type: geometry type to be kept.
        n_jobs: Number of treads.
    """
    geoms = GeoSeries(geoms)
    geoms.index = range(len(geoms))
    geoms.loc[:] = make_valid(geoms.to_numpy())
    geoms = geoms.explode(index_parts=False).pipe(to_single_geom_type, geom_type)
    only_one = geoms.groupby(level=0).transform("size") == 1
    one_hit = geoms[only_one]
    many_hits = geoms[~only_one].groupby(level=0).agg(unary_union)
    return pd.concat([one_hit, many_hits]).sort_index()


def _agg_geoms(g: np.ndarray, grid_size: int | float | None = None) -> Geometry:
    return make_valid(unary_union(g, grid_size=grid_size))



import functools
import numbers
import os
import re
import warnings
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from copy import copy
from copy import deepcopy
from json import loads
from pathlib import Path
from typing import Any
from typing import ClassVar

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj
import rasterio
import rasterio.windows
import shapely
from typing_extensions import Self  # TODO: imperter fra typing når python 3.11

try:
    import xarray as xr
    from xarray import DataArray
except ImportError:

    class DataArray:
        """Placeholder."""


try:
    from dapla.gcs import GCSFileSystem
except ImportError:

    class GCSFileSystem:
        """Placeholder."""


try:
    from rioxarray.rioxarray import _generate_spatial_coords
except ImportError:
    pass
from affine import Affine
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from pandas.api.types import is_list_like
from rasterio import features
from rasterio.enums import MergeAlg
from rasterio.io import DatasetReader
from rasterio.vrt import WarpedVRT
from rasterio.warp import reproject
from shapely import Geometry
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import shape

from ..geopandas_tools.conversion import to_bbox
from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.conversion import to_shapely
from ..geopandas_tools.general import is_bbox_like
from ..geopandas_tools.general import is_wkt
from ..helpers import is_property
from ..io.opener import opener
from .base import ALLOWED_KEYS
from .base import NESSECARY_META
from .base import get_index_mapper
from .base import memfile_from_array
from .zonal import _aggregate
from .zonal import _make_geometry_iterrows
from .zonal import _no_overlap_df
from .zonal import _prepare_zonal
from .zonal import _zonal_post

numpy_func_message = (
    "aggfunc must be functions or strings of numpy functions or methods."
)


class Raster:
    """For reading, writing and working with rasters.

    Raster instances should be created with the methods 'from_path', 'from_array' or
    'from_gdf'.


    Examples:
    ---------
    Read tif file.

    >>> import sgis as sg
    >>> path = 'https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/raster/dtm_10.tif'
    >>> raster = sg.Raster.from_path(path)
    >>> raster
    Raster(shape=(1, 201, 201), res=10, crs=ETRS89 / UTM zone 33N (N-E), path=https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/raster/dtm_10.tif)

    Load the entire image as an numpy ndarray.
    Operations are done in place to save memory.
    The array is stored in the array attribute.

    >>> raster.load()
    >>> raster.values[raster.values < 0] = 0
    >>> raster.values
    [[[  0.    0.    0.  ... 158.4 155.6 152.6]
    [  0.    0.    0.  ... 158.  154.8 151.9]
    [  0.    0.    0.  ... 158.5 155.1 152.3]
    ...
    [  0.  150.2 150.6 ...   0.    0.    0. ]
    [  0.  149.9 150.1 ...   0.    0.    0. ]
    [  0.  149.2 149.5 ...   0.    0.    0. ]]]

    Save as tif file.

    >>> raster.write("path/to/file.tif")

    Convert to GeoDataFrame.

    >>> gdf = raster.to_gdf(column="elevation")
    >>> gdf
           elevation                                           geometry  indexes
    0            1.9  POLYGON ((-25665.000 6676005.000, -25665.000 6...           1
    1           11.0  POLYGON ((-25655.000 6676005.000, -25655.000 6...           1
    2           18.1  POLYGON ((-25645.000 6676005.000, -25645.000 6...           1
    3           15.8  POLYGON ((-25635.000 6676005.000, -25635.000 6...           1
    4           11.6  POLYGON ((-25625.000 6676005.000, -25625.000 6...           1
    ...          ...                                                ...         ...
    25096       13.4  POLYGON ((-24935.000 6674005.000, -24935.000 6...           1
    25097        9.4  POLYGON ((-24925.000 6674005.000, -24925.000 6...           1
    25098        5.3  POLYGON ((-24915.000 6674005.000, -24915.000 6...           1
    25099        2.3  POLYGON ((-24905.000 6674005.000, -24905.000 6...           1
    25100        0.1  POLYGON ((-24895.000 6674005.000, -24895.000 6...           1

    The image can also be clipped by a mask while loading.

    >>> small_circle = raster_as_polygons.unary_union.centroid.buffer(50)
    >>> raster = sg.Raster.from_path(path).clip(small_circle)
    Raster(shape=(1, 10, 10), res=10, crs=ETRS89 / UTM zone 33N (N-E), path=https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/raster/dtm_10.tif)

    Construct raster from GeoDataFrame.
    The arrays are put on top of each other in a 3 dimensional array.

    >>> raster_as_polygons["elevation_x2"] = raster_as_polygons["elevation"] * 2
    >>> raster_from_polygons = sg.Raster.from_gdf(raster_as_polygons, columns=["elevation", "elevation_x2"], res=20)
    >>> raster_from_polygons
    Raster(shape=(2, 100, 100), res=20, raster_id=-260056673995, crs=ETRS89 / UTM zone 33N (N-E), path=None)

    Calculate zonal statistics for each polygon in 'gdf'.

    >>> zonal = raster.zonal(raster_as_polygons, aggfunc=["sum", np.mean])
    >>> zonal
            sum  mean                                           geometry
    0       1.9   1.9  POLYGON ((-25665.000 6676005.000, -25665.000 6...
    1      11.0  11.0  POLYGON ((-25655.000 6676005.000, -25655.000 6...
    2      18.1  18.1  POLYGON ((-25645.000 6676005.000, -25645.000 6...
    3      15.8  15.8  POLYGON ((-25635.000 6676005.000, -25635.000 6...
    4      11.6  11.6  POLYGON ((-25625.000 6676005.000, -25625.000 6...
    ...     ...   ...                                                ...
    25096  13.4  13.4  POLYGON ((-24935.000 6674005.000, -24935.000 6...
    25097   9.4   9.4  POLYGON ((-24925.000 6674005.000, -24925.000 6...
    25098   5.3   5.3  POLYGON ((-24915.000 6674005.000, -24915.000 6...
    25099   2.3   2.3  POLYGON ((-24905.000 6674005.000, -24905.000 6...
    25100   0.1   0.1  POLYGON ((-24895.000 6674005.000, -24895.000 6...

    """

    # attributes concerning rasterio metadata
    _profile: ClassVar[dict[str, str | None]] = {
        "driver": "GTiff",
        "compress": "LZW",
        "nodata": None,
        "dtype": None,
        "crs": None,
        "tiled": None,
        "indexes": None,
    }

    def __init__(
        self,
        data: Self | str | np.ndarray | None = None,
        *,
        file_system: GCSFileSystem | None = None,
        filename_regex: str | None = None,
        **kwargs,
    ) -> None:
        """Note: use the classmethods from_path, from_array, from_gdf etc. instead of the initialiser.

        Args:
            data: A file path, an array or a Raster object.
            file_system: Optional GCSFileSystem.
            filename_regex: Regular expression to match file name attributes (date, band, tile, resolution).
            **kwargs: Arguments concerning file metadata or
                spatial properties of the image.
        """
        warnings.warn("This class is deprecated in favor of Band", stacklevel=1)
        self.filename_regex = filename_regex
        if filename_regex:
            self.filename_pattern = re.compile(self.filename_regex, re.VERBOSE)
        else:
            self.filename_pattern = None

        if isinstance(data, Raster):
            for key, value in data.__dict__.items():
                setattr(data, key, value)
            return

        if isinstance(data, (str | Path | os.PathLike)):
            self.path = data

        else:
            self.path = None

        if isinstance(data, (np.ndarray)):
            self.values = data
        else:
            self.values = None

        if self.path is None and not any(
            [kwargs.get("transform"), kwargs.get("bounds")]
        ):
            raise TypeError(
                "Must specify either bounds or transform when constructing raster from array."
            )

        # add class profile first, then override with args and kwargs
        self.update(**self._profile)

        self._crs = kwargs.pop("crs", self._crs if hasattr(self, "_crs") else None)
        self._bounds = None
        self.file_system = file_system
        self._indexes = self._get_indexes(kwargs.pop("indexes", self.indexes))

        # override the above with kwargs
        self.update(**kwargs)

        attributes = set(self.__dict__.keys()).difference(set(self.properties))

        if self.path is not None and not self._has_nessecary_attrs(attributes):
            self._add_meta()
            self._meta_added = True

        self._prev_crs = self._crs

    @classmethod
    def from_path(
        cls,
        path: str,
        res: int | None = None,
        file_system: GCSFileSystem | None = None,
        filename_regex: str | None = None,
        **kwargs,
    ) -> Self:
        """Construct Raster from file path.

        Args:
            path: Path to a raster image file.
            res: Spatial resolution when reading the image.
            file_system: Optional file system.
            filename_regex: Regular expression with optional match groups.
            **kwargs: Arguments concerning file metadata or
                spatial properties of the image.

        Returns:
            A Raster instance.
        """
        return cls(
            str(path),
            file_system=file_system,
            res=res,
            filename_regex=filename_regex,
            **kwargs,
        )

    @classmethod
    def from_array(
        cls,
        array: np.ndarray,
        crs: Any,
        *,
        transform: Affine | None = None,
        bounds: tuple | Geometry | None = None,
        copy: bool = True,
        **kwargs,
    ) -> Self:
        """Construct Raster from numpy array.

        Must also specify nessecary spatial properties
        The necessary metadata is 'crs' and either 'transform' (Affine object)
        or 'bounds', which transform will then be created from.

        Args:
            array: 2d or 3d numpy ndarray.
            crs: Coordinate reference system.
            transform: Affine transform object. Can be specified instead
                of bounds.
            bounds: Minimum and maximum x and y coordinates. Can be specified instead
                of transform.
            copy: Whether to copy the array.
            **kwargs: Arguments concerning file metadata or
                spatial properties of the image.

        Returns:
            A Raster instance.
        """
        if array is None:
            raise TypeError("Must specify array.")

        if not any([transform, bounds]):
            raise TypeError(
                "Must specify either bounds or transform when constructing raster from array."
            )

        array = array.copy() if copy else array

        if len(array.shape) == 2:
            height, width = array.shape
        elif len(array.shape) == 3:
            height, width = array.shape[1:]
        else:
            raise ValueError("array must be 2 or 3 dimensional.")

        transform = Affine(*transform) if transform is not None else None

        if bounds is not None:
            bounds = to_bbox(bounds)

        if transform and not bounds:
            bounds = rasterio.transform.array_bounds(height, width, transform)

        crs = pyproj.CRS(crs) if crs else None

        return cls(array, crs=crs, transform=transform, bounds=bounds, **kwargs)

    @classmethod
    def from_gdf(
        cls,
        gdf: GeoDataFrame,
        columns: str | Iterable[str],
        res: int,
        fill: int = 0,
        all_touched: bool = False,
        merge_alg: Callable = MergeAlg.replace,
        default_value: int = 1,
        dtype: Any | None = None,
        **kwargs,
    ) -> Self:
        """Construct Raster from a GeoDataFrame.

        Args:
            gdf: The GeoDataFrame to rasterize.
            columns: Column(s) in the GeoDataFrame whose values are used to populate the raster.
                This can be a single column name or a list of column names.
            res: Resolution of the raster in units of the GeoDataFrame's coordinate reference system.
            fill: Fill value for areas outside of input geometries (default is 0).
            all_touched: Whether to consider all pixels touched by geometries,
                not just those whose center is within the polygon (default is False).
            merge_alg: Merge algorithm to use when combining geometries
                (default is 'MergeAlg.replace').
            default_value: Default value to use for the rasterized pixels
                (default is 1).
            dtype: Data type of the output array. If None, it will be
                determined automatically.
            **kwargs: Additional keyword arguments passed to the raster
                creation process, e.g., custom CRS or transform settings.

        Returns:
            A Raster instance based on the specified GeoDataFrame and parameters.

        Raises:
            TypeError: If 'transform' is provided in kwargs, as this is
            computed based on the GeoDataFrame bounds and resolution.
        """
        if not isinstance(gdf, GeoDataFrame):
            gdf = to_gdf(gdf)

        if "transform" in kwargs:
            raise TypeError("Unexpected argument 'transform'")

        kwargs["crs"] = gdf.crs or kwargs.get("crs")

        if kwargs["crs"] is None:
            raise TypeError("Must specify crs if the object doesn't have crs.")

        shape = get_shape_from_bounds(gdf.total_bounds, res=res)
        transform = get_transform_from_bounds(gdf.total_bounds, shape)
        kwargs["transform"] = transform

        def _rasterize(gdf, col):
            return features.rasterize(
                cls._gdf_to_geojson_with_col(gdf, col),
                out_shape=shape,
                transform=transform,
                fill=fill,
                all_touched=all_touched,
                merge_alg=merge_alg,
                default_value=default_value,
                dtype=dtype,
            )

        # make 2d array
        if isinstance(columns, str):
            array = _rasterize(gdf, columns)
            assert len(array.shape) == 2
            name = kwargs.get("name", columns)

        # 3d array even if single column in list/tuple
        elif hasattr(columns, "__iter__"):
            array = []
            for col in columns:
                arr = _rasterize(gdf, col)
                array.append(arr)
            array = np.array(array)
            assert len(array.shape) == 3
            name = kwargs.get("name", None)

        return cls.from_array(array, name=name, **kwargs)

    @classmethod
    def from_dict(cls, dictionary: dict) -> Self:
        """Construct Raster from metadata dict to fastpass the initializer.

        This is the fastest way to create a Raster since a metadata lookup is not
        needed.

        The dictionary must have all the keys ...
        and at least one of the keys 'transform' and 'bounds'.

        Args:
            dictionary: Dictionary with the nessecary and optional information
                about the raster. This can be fetched from an existing raster with
                the to_dict method.

        Returns:
            A Raster instance.
        """
        cls._validate_dict(dictionary)

        return cls(**dictionary)

    def update(self, **kwargs) -> Self:
        """Update attributes of the Raster."""
        for key, value in kwargs.items():
            self._validate_key(key)
            if is_property(self, key):
                key = "_" + key
            setattr(self, key, value)
        return self

    def write(
        self, path: str, window: rasterio.windows.Window | None = None, **kwargs
    ) -> None:
        """Write the raster as a single file.

        Multiband arrays will result in a multiband image file.

        Args:
            path: File path to write to.
            window: Optional window to clip the image to.
            **kwargs: Keyword arguments passed to rasterio.open.
                Thise will override the items in the Raster's profile,
                if overlapping.
        """
        if self.values is None:
            raise AttributeError("The image hasn't been loaded.")

        profile = self.profile | kwargs

        with opener(path, "wb", file_system=self.file_system) as file:
            with rasterio.open(file, "w", **profile) as dst:
                self._write(dst, window)

        self.path = str(path)

    def load(self, reload: bool = False, **kwargs) -> Self:
        """Load the entire image as an np.array.

        The array is stored in the 'array' attribute
        of the Raster.

        Args:
            reload: Whether to reload the array if already loaded.
            **kwargs: Keyword arguments passed to the rasterio read
                method.
        """
        if "mask" in kwargs:
            raise ValueError("Got an unexpected keyword argument 'mask'")
        if "window" in kwargs:
            raise ValueError("Got an unexpected keyword argument 'window'")

        if reload or self.values is None:
            self._read_tif(**kwargs)

        return self

    def clip(
        self,
        mask: Any,
        masked: bool = False,
        boundless: bool = True,
        **kwargs,
    ) -> Self:
        """Load the part of the image inside the mask.

        The returned array is stored in the 'array' attribute
        of the Raster.

        Args:
            mask: Geometry-like object or bounding box.
            masked: If 'masked' is True the return value will be a masked
                array. Otherwise (default) the return value will be a
                regular array. Masks will be exactly the inverse of the
                GDAL RFC 15 conforming arrays returned by read_masks().
            boundless: If True, windows that extend beyond the dataset's extent
                are permitted and partially or completely filled arrays will
                be returned as appropriate.
            **kwargs: Keyword arguments passed to the mask function
                from the rasterio.mask module.

        Returns:
            Self, but with the array loaded.
        """
        if not isinstance(mask, GeoDataFrame):
            mask = self._return_gdf(mask)

        try:
            mask = mask.to_crs(self.crs)
        except ValueError:
            mask = mask.set_crs(self.crs)

        self._read_with_mask(mask=mask, masked=masked, boundless=boundless, **kwargs)

        return self

    def intersects(self, other: Any) -> bool:
        """Returns True if the image bounds intersect with 'other'."""
        return self.unary_union.intersects(to_shapely(other))

    def sample(
        self, n: int = 1, size: int = 20, mask: Any = None, copy: bool = True, **kwargs
    ) -> Self:
        """Take a random spatial sample of the image."""
        if mask is not None:
            points = GeoSeries(self.unary_union).clip(mask).sample_points(n)
        else:
            points = GeoSeries(self.unary_union).sample_points(n)
        buffered = points.buffer(size / self.res)
        boxes = to_gdf(
            [shapely.box(*arr) for arr in buffered.bounds.values], crs=self.crs
        )
        if copy:
            copy = self.copy()
            return copy.clip(boxes, **kwargs)
        return self.clip(boxes, **kwargs)

    def zonal(
        self,
        polygons: GeoDataFrame,
        aggfunc: str | Callable | list[Callable | str],
        array_func: Callable | None = None,
        dropna: bool = True,
    ) -> GeoDataFrame:
        """Calculate zonal statistics in polygons.

        Args:
            polygons: A GeoDataFrame of polygon geometries.
            aggfunc: Function(s) of which to aggregate the values
                within each polygon.
            array_func: Optional calculation of the raster
                array before calculating the zonal statistics.
            dropna: If True (default), polygons with all missing
                values will be removed.

        Returns:
            A GeoDataFrame with aggregated values per polygon.
        """
        idx_mapper, idx_name = get_index_mapper(polygons)
        polygons, aggfunc, func_names = _prepare_zonal(polygons, aggfunc)
        poly_iter = _make_geometry_iterrows(polygons)

        aggregated = []
        for i, poly in poly_iter:
            clipped = self.clip(poly)
            if not np.size(clipped.values):
                aggregated.append(_no_overlap_df(func_names, i, date=self.date))
            aggregated.append(
                _aggregate(
                    clipped.values, array_func, aggfunc, func_names, self.date, i
                )
            )

        return _zonal_post(
            aggregated,
            polygons=polygons,
            idx_mapper=idx_mapper,
            idx_name=idx_name,
            dropna=dropna,
        )

    def to_xarray(self) -> DataArray:
        """Convert the raster to  an xarray.DataArray."""
        self._check_for_array()
        self.name = self.name or self.__class__.__name__.lower()
        coords = _generate_spatial_coords(self.transform, self.width, self.height)
        if len(self.values.shape) == 2:
            dims = ["y", "x"]
            # dims = ["band", "y", "x"]
            # array = np.array([self.values])
            # assert len(array.shape) == 3
        elif len(self.values.shape) == 3:
            dims = ["band", "y", "x"]
            # array = self.values
        else:
            raise ValueError("Array must be 2 or 3 dimensional.")
        return xr.DataArray(
            self.values,
            coords=coords,
            dims=dims,
            name=self.name,
            attrs={"crs": self.crs},
        )  # .transpose("y", "x")

    def to_dict(self) -> dict:
        """Get a dictionary of Raster attributes."""
        out = {}
        for col in self.ALL_ATTRS:
            try:
                out[col] = self[col]
            except AttributeError:
                pass
        return out

    def to_gdf(self, column: str | list[str] | None = None) -> GeoDataFrame:
        """Create a GeoDataFrame from the raster.

        For multiband rasters, the bands are in separate rows with a "band" column
        value corresponding to the band indexes of the raster.

        Args:
            column: Name of resulting column(s) that holds the raster values.
                Can be a single string or an iterable with the same length as
                the number of raster bands.

        Returns:
            A GeoDataFrame with a geometry column, a 'band' column and a
            one or more value columns.
        """
        self._check_for_array()

        array_list = self.array_list()

        if is_list_like(column) and len(column) != len(array_list):
            raise ValueError(
                "columns should be a string or a list of same length as "
                f"layers in the array ({len(array_list)})."
            )

        if column is None:
            column = ["value"] * len(array_list)

        if isinstance(column, str):
            column = [column] * len(array_list)

        gdfs = []
        for i, (col, array) in enumerate(zip(column, array_list, strict=True)):
            gdf = gpd.GeoDataFrame(
                pd.DataFrame(
                    self._array_to_geojson(array, self.transform),
                    columns=[col, "geometry"],
                ),
                geometry="geometry",
                crs=self.crs,
            )
            gdf["indexes"] = i + 1
            gdfs.append(gdf)

        return pd.concat(gdfs, ignore_index=True)

    def set_crs(
        self,
        crs: pyproj.CRS | Any,
        allow_override: bool = False,
    ) -> Self:
        """Set coordinate reference system."""
        if not allow_override and self.crs is not None:
            raise ValueError("Cannot overwrite crs when allow_override is False.")

        if self.values is None:
            raise ValueError("array must be loaded/clipped before set_crs")

        self._crs = pyproj.CRS(crs)
        return self

    def to_crs(self, crs: pyproj.CRS | Any, **kwargs) -> Self:
        """Reproject the raster.

        Args:
            crs: The new coordinate reference system.
            **kwargs: Keyword arguments passed to the reproject function
                from the rasterio.warp module.
        """
        if self.crs is None:
            raise ValueError("Raster has no crs. Use set_crs.")

        # if pyproj.CRS(crs).equals(pyproj.CRS(self._crs)) and pyproj.CRS(crs).equals(
        #     pyproj.CRS(self._prev_crs)
        # ):
        #     return self

        if self.values is None:
            project = pyproj.Transformer.from_crs(
                pyproj.CRS(self._prev_crs), pyproj.CRS(crs), always_xy=True
            ).transform

            old_box = shapely.box(*self.bounds)
            new_box = shapely.ops.transform(project, old_box)
            self._bounds = to_bbox(new_box)

            # TODO: fix area changing... if possible
            # print("old/new:", shapely.area(old_box) / shapely.area(new_box))

            if pyproj.CRS(crs).equals(pyproj.CRS(self._crs)):
                self._warped_crs = self._crs
                return self

            # self._bounds = rasterio.warp.transform_bounds(
            #     pyproj.CRS(self._prev_crs), pyproj.CRS(crs), *to_bbox(self._bounds)
            # )
            # transformer = pyproj.Transformer.from_crs(
            #     pyproj.CRS(self._prev_crs), pyproj.CRS(crs), always_xy=True
            # )
            # minx, miny, maxx, maxy = self.bounds
            # xs, ys = transformer.transform(xx=[minx, maxx], yy=[miny, maxy])

            # minx, maxx = xs
            # miny, maxy = ys
            # self._bounds = minx, miny, maxx, maxy

            # self._bounds = shapely.transform(old_box, project)
        else:
            was_2d = len(self.shape) == 2
            self.values, transform = reproject(
                source=self.values,
                src_crs=self._prev_crs,
                src_transform=self.transform,
                dst_crs=pyproj.CRS(crs),
                **kwargs,
            )
            if was_2d and len(self.values.shape) == 3:
                assert self.values.shape[0] == 1
                self.values = self.values[0]

            self._bounds = rasterio.transform.array_bounds(
                self.height, self.width, transform
            )

        self._warped_crs = pyproj.CRS(crs)
        self._prev_crs = pyproj.CRS(crs)

        return self

    def plot(self, mask: Any | None = None) -> None:
        """Plot the images. One image per band."""
        self._check_for_array()
        if mask is not None:
            raster = self.copy().clip(mask)
        else:
            raster = self

        if len(raster.shape) == 2:
            array = np.array([raster.values])
        else:
            array = raster.values

        for arr in array:
            ax = plt.axes()
            ax.imshow(arr)
            ax.axis("off")
            plt.show()
            plt.close()

    def astype(self, dtype: type) -> Self:
        """Convert the datatype of the array."""
        if self.values is None:
            raise ValueError("Array is not loaded.")
        if not rasterio.dtypes.can_cast_dtype(self.values, dtype):
            min_dtype = rasterio.dtypes.get_minimum_dtype(self.values)
            raise ValueError(f"Cannot cast to dtype. Minimum dtype is {min_dtype}")
        self.values = self.values.astype(dtype)
        self._dtype = dtype
        return self

    def as_minimum_dtype(self) -> Self:
        """Convert the array to the minimum dtype without overflow."""
        min_dtype = rasterio.dtypes.get_minimum_dtype(self.values)
        self.values = self.values.astype(min_dtype)
        return self

    def min(self) -> int | None:
        """Minimum value in the array."""
        if np.size(self.values):
            return np.min(self.values)
        return None

    def max(self) -> int | None:
        """Maximum value in the array."""
        if np.size(self.values):
            return np.max(self.values)
        return None

    def _add_meta(self) -> Self:
        mess = "Cannot add metadata after image has been "
        if hasattr(self, "_clipped"):
            raise ValueError(mess + "clipped.")
        if hasattr(self, "_warped_crs"):
            raise ValueError(mess + "reprojected.")

        with opener(self.path, file_system=self.file_system) as file:
            with rasterio.open(file) as src:
                self._add_meta_from_src(src)

        return self

    def array_list(self) -> list[np.ndarray]:
        """Get a list of 2D arrays."""
        self._check_for_array()
        if len(self.values.shape) == 2:
            return [self.values]
        elif len(self.values.shape) == 3:
            return list(self.values)
        else:
            raise ValueError

    @property
    def indexes(self) -> int | tuple[int] | None:
        """Band indexes of the image."""
        return self._indexes

    @property
    def name(self) -> str | None:
        """Name of the file in the file path, if any."""
        try:
            return self._name
        except AttributeError:
            try:
                return Path(self.path).name
            except TypeError:
                return None

    @name.setter
    def name(self, value) -> None:
        self._name = value

    @property
    def date(self) -> str | None:
        """Date in the image file name, if filename_regex is present."""
        try:
            return re.match(self.filename_pattern, Path(self.path).name).group("date")
        except (AttributeError, TypeError):
            return None

    @property
    def band(self) -> str | None:
        """Band name of the image file name, if filename_regex is present."""
        try:
            return re.match(self.filename_pattern, Path(self.path).name).group("band")
        except (AttributeError, TypeError):
            return None

    @property
    def dtype(self) -> Any:
        """Data type of the array."""
        try:
            return self.values.dtype
        except AttributeError:
            try:
                return self._dtype
            except AttributeError:
                return None

    @dtype.setter
    def dtype(self, new_dtype: Any) -> None:
        self.values = self.values.astype(new_dtype)

    @property
    def nodata(self) -> int | None:
        """No data value."""
        try:
            return self._nodata
        except AttributeError:
            return None

    @property
    def tile(self) -> str | None:
        """Tile name from regex."""
        try:
            return re.match(self.filename_pattern, Path(self.path).name).group("tile")
        except (AttributeError, TypeError):
            return None

    @property
    def meta(self) -> dict:
        """Metadata dict."""
        return {
            "path": self.path,
            "type": self.__class__.__name__,
            "bounds": self.bounds,
            "indexes": self.indexes,
            "crs": self.crs,
        }

    @property
    def profile(self) -> dict:
        """Profile of the image file."""
        # TODO: .crs blir feil hvis warpa. Eller?
        return {
            "driver": self.driver,
            "compress": self.compress,
            "dtype": self.dtype,
            "crs": self.crs,
            "transform": self.transform,
            "nodata": self.nodata,
            "count": self.count,
            "height": self.height,
            "width": self.width,
            "indexes": self.indexes,
        }

    @property
    def read_kwargs(self) -> dict:
        """Keywords passed to the read method of rasterio.io.DatasetReader."""
        return {
            "indexes": self.indexes,
            "fill_value": self.nodata,
            "masked": False,
        }

    @property
    def res(self) -> float | None:
        """Get the spatial resolution of the image."""
        if hasattr(self, "_res") and self._res is not None:
            return self._res
        if self.width is None:
            return None
        diffx = self.bounds[2] - self.bounds[0]
        return diffx / self.width

    @property
    def height(self) -> int | None:
        """Get the height of the image as number of pixels."""
        if self.values is None:
            try:
                return self._height
            except AttributeError:
                return None
        i = 1 if len(self.values.shape) == 3 else 0
        return self.values.shape[i]

    @property
    def width(self) -> int | None:
        """Get the width of the image as number of pixels."""
        if self.values is None:
            try:
                return self._width
            except AttributeError:
                try:
                    heigth, width = get_shape_from_bounds(self, self.res)  # .res[0])
                    self._width = width
                    self._heigth = heigth
                    return self._width
                except Exception:
                    return None
        i = 2 if len(self.values.shape) == 3 else 1
        return self.values.shape[i]

    @property
    def count(self) -> int:
        """Get the number of bands in the image."""
        if self.values is not None:
            if len(self.values.shape) == 3:
                return self.values.shape[0]
            if len(self.values.shape) == 2:
                return 1
        if not hasattr(self._indexes, "__iter__"):
            return 1
        return len(self._indexes)

    @property
    def shape(self) -> tuple[int]:
        """Shape that is consistent with the array, whether it is loaded or not."""
        if self.values is not None:
            return self.values.shape
        if hasattr(self._indexes, "__iter__"):
            return self.count, self.width, self.height
        return self.width, self.height

    @property
    def transform(self) -> Affine | None:
        """Get the Affine transform of the image."""
        try:
            return rasterio.transform.from_bounds(*self.bounds, self.width, self.height)
        except (ZeroDivisionError, TypeError):
            if not self.width or not self.height:
                return None

    @property
    def bounds(self) -> tuple[float, float, float, float] | None:
        """Get the bounds of the image."""
        try:
            return to_bbox(self._bounds)
        except (AttributeError, TypeError):
            return None

    @property
    def crs(self) -> pyproj.CRS | None:
        """Get the coordinate reference system of the image."""
        try:
            return self._warped_crs
        except AttributeError:
            try:
                return self._crs
            except AttributeError:
                return None

    @property
    def area(self) -> float:
        """Get the area of the image."""
        return shapely.area(self.unary_union)

    @property
    def length(self) -> float:
        """Get the circumfence of the image."""
        return shapely.length(self.unary_union)

    @property
    def unary_union(self) -> Polygon:
        """Get the image bounds as a Polygon."""
        return shapely.box(*self.bounds)

    @property
    def centroid(self) -> Point:
        """Get the centerpoint of the image."""
        x = (self.bounds[0] + self.bounds[2]) / 2
        y = (self.bounds[1] + self.bounds[3]) / 2
        return Point(x, y)

    @property
    def properties(self) -> list[str]:
        """List of all properties of the class."""
        out = []
        for attr in dir(self):
            try:
                if is_property(self, attr):
                    out.append(attr)
            except AttributeError:
                pass
        return out

    def indexes_as_tuple(self) -> tuple[int, ...]:
        """Get the band index(es) as a tuple of integers."""
        if len(self.shape) == 2:
            return (1,)
        return tuple(i + 1 for i in range(self.shape[0]))

    def copy(self, deep: bool = True) -> "Raster":
        """Returns a (deep) copy of the class instance.

        Args:
            deep: Whether to return a deep or shallow copy. Defaults to True.
        """
        if deep:
            return deepcopy(self)
        else:
            return copy(self)

    def equals(self, other: Any) -> bool:
        """Check if the Raster is equal to another Raster."""
        if not isinstance(other, Raster):
            raise NotImplementedError("other must be of type Raster")
        if type(other) != type(self):
            return False
        if self.values is None and other.values is not None:
            return False
        if self.values is not None and other.values is None:
            return False

        for method in dir(self):
            if not is_property(self, method):
                continue
            if getattr(self, method) != getattr(other, method):
                return False

        return np.array_equal(self.values, other.values)

    def __repr__(self) -> str:
        """The print representation."""
        shape = self.shape
        shp = ", ".join([str(x) for x in shape])
        try:
            res = int(self.res)
        except TypeError:
            res = None
        return f"{self.__class__.__name__}(shape=({shp}), res={res}, band={self.band})"

    def __iter__(self) -> Iterator[np.ndarray]:
        """Iterate over the arrays."""
        if len(self.values.shape) == 2:
            return iter([self.values])
        if len(self.values.shape) == 3:
            return iter(self.values)
        raise ValueError(
            f"Array should have shape length 2 or 3. Got {len(self.values.shape)}"
        )

    def __mul__(self, scalar: int | float) -> "Raster":
        """Multiply the array values with *."""
        self._check_for_array()
        self.values = self.values * scalar
        return self

    def __add__(self, scalar: int | float) -> "Raster":
        """Add to the array values with +."""
        self._check_for_array()
        self.values = self.values + scalar
        return self

    def __sub__(self, scalar: int | float) -> "Raster":
        """Subtract the array values with -."""
        self._check_for_array()
        self.values = self.values - scalar
        return self

    def __truediv__(self, scalar: int | float) -> "Raster":
        """Divide the array values with /."""
        self._check_for_array()
        self.values = self.values / scalar
        return self

    def __floordiv__(self, scalar: int | float) -> "Raster":
        """Floor divide the array values with //."""
        self._check_for_array()
        self.values = self.values // scalar
        return self

    def __pow__(self, exponent: int | float) -> "Raster":
        """Exponentiate the array values with **."""
        self._check_for_array()
        self.values = self.values**exponent
        return self

    def _has_nessecary_attrs(self, dict_like: dict) -> bool:
        """Check if Raster init got enough kwargs to not need to read src."""
        try:
            self._validate_dict(dict_like)
            return all(
                x is not None for x in [self.indexes, self.res, self.crs, self.bounds]
            )
        except AttributeError:
            return False

    def _return_self_or_copy(self, array: np.ndarray, copy: bool) -> "Raster":
        if not copy:
            self.values = array
            return self
        else:
            copy = self.copy()
            copy.values = array
            return copy

    @classmethod
    def _validate_dict(cls, dict_like: dict) -> None:
        missing = []
        for attr in NESSECARY_META:
            if any(
                [
                    attr in dict_like,
                    f"_{attr}" in dict_like,
                    attr.lstrip("_") in dict_like,
                ]
            ):
                continue
            missing.append(attr)
        if missing:
            raise AttributeError(f"Missing nessecary key(s) {', '.join(missing)}")

    @classmethod
    def _validate_key(cls, key: str) -> None:
        if key not in ALLOWED_KEYS:
            raise ValueError(
                f"Got an unexpected key {key!r}. Allowed keys are ",
                ", ".join(ALLOWED_KEYS),
            )

    def _get_shape_from_res(self, res: int) -> tuple[int] | None:
        if res is None:
            return None
        if hasattr(res, "__iter__") and len(res) == 2:
            res = res[0]
        diffx = self.bounds[2] - self.bounds[0]
        diffy = self.bounds[3] - self.bounds[1]
        width = int(diffx / res)
        height = int(diffy / res)
        if hasattr(self.indexes, "__iter__"):
            return len(self.indexes), width, height
        return width, height

    def _write(
        self, dst: rasterio.io.DatasetReader, window: rasterio.windows.Window
    ) -> None:
        if np.ma.is_masked(self.values):
            if len(self.values.shape) == 2:
                return dst.write(
                    self.values.filled(self.nodata), indexes=1, window=window
                )

            for i in range(len(self.indexes_as_tuple())):
                dst.write(
                    self.values[i].filled(self.nodata),
                    indexes=i + 1,
                    window=window,
                )

        else:
            if len(self.values.shape) == 2:
                return dst.write(self.values, indexes=1, window=window)

            for i, idx in enumerate(self.indexes_as_tuple()):
                dst.write(self.values[i], indexes=idx, window=window)

    def _get_indexes(self, indexes: int | tuple[int] | None) -> int | tuple[int] | None:
        if isinstance(indexes, numbers.Number):
            return int(indexes)
        if indexes is None:
            if self.values is not None and len(self.values.shape) == 3:
                return tuple(i + 1 for i in range(self.values.shape[0]))
            elif self.values is not None and len(self.values.shape) == 2:
                return 1
            elif self.values is not None:
                raise ValueError("Array must be 2 or 3 dimensional.")
            else:
                return None
        try:
            return tuple(int(x) for x in indexes)
        except Exception as e:
            raise TypeError(
                "indexes should be an integer or an iterable of integers."
                f"Got {type(indexes)}: {indexes}"
            ) from e

    def _return_gdf(self, obj: Any) -> GeoDataFrame:
        if isinstance(obj, str) and not is_wkt(obj):
            return self._read_tif(obj)
        elif isinstance(obj, Raster):
            return obj.to_gdf()
        elif is_bbox_like(obj):
            return to_gdf(shapely.box(*to_bbox(obj)), crs=self.crs)
        else:
            return to_gdf(obj, crs=self.crs)

    @staticmethod
    def _gdf_to_geojson(gdf: GeoDataFrame) -> list[dict]:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            return [x["geometry"] for x in loads(gdf.to_json())["features"]]

    @staticmethod
    def _gdf_to_geojson_with_col(gdf: GeoDataFrame, column: str) -> list[dict]:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            return [
                (feature["geometry"], val)
                for val, feature in zip(
                    gdf[column], loads(gdf.to_json())["features"], strict=False
                )
            ]

    @staticmethod
    def _array_to_geojson(array: np.ndarray, transform: Affine) -> list[tuple]:
        if np.ma.is_masked(array):
            array = array.data
        try:
            return [
                (value, shape(geom))
                for geom, value in features.shapes(
                    array, transform=transform, mask=None
                )
            ]
        except ValueError:
            array = array.astype(np.float32)
            return [
                (value, shape(geom))
                for geom, value in features.shapes(
                    array, transform=transform, mask=None
                )
            ]

    def _add_indexes_from_array(self, indexes: int | tuple[int]) -> int | tuple[int]:
        if indexes is not None:
            return indexes
        elif len(self.values.shape) == 3:
            return tuple(x + 1 for x in range(len(self.values)))
        elif len(self.values.shape) == 2:
            return 1
        else:
            raise ValueError

    def _add_meta_from_src(self, src: rasterio.io.DatasetReader) -> None:
        if not hasattr(self, "_bounds") or self._bounds is None:
            self._bounds = tuple(src.bounds)

        try:
            self._crs = pyproj.CRS(src.crs)
        except pyproj.exceptions.CRSError:
            self._crs = None

        self._width = src.width
        self._height = src.height

        # for attr in dir(self):
        #     try:
        #         if is_property(self, attr):
        #             continue
        #         if attr is None:
        #             new_value = getattr(src, attr)
        #             setattr(self, attr, new_value)
        #     except AttributeError:
        #         pass

        if not hasattr(self, "_indexes") or self._indexes is None:
            new_value = src.indexes
            if new_value == 1 or new_value == (1,):
                new_value = 1
            self._indexes = new_value

        if not hasattr(self, "_nodata") or self._nodata is None:
            new_value = src.nodata
            self._nodata = new_value

        # if not hasattr(self, "_indexes") or self._indexes is None:
        #     self._indexes = src.indexes

        # if not hasattr(self, "_nodata") or self._nodata is None:
        #     self._nodata = src.nodata

    def _load_warp_file(self) -> DatasetReader:
        """(from Torchgeo). Load and warp a file to the correct CRS and resolution.

        Args:
            filepath: file to load and warp

        Returns:
            file handle of warped VRT
        """
        with opener(self.path, file_system=self.file_system) as file:
            src = rasterio.open(file)

        # Only warp if necessary
        if src.crs != self.crs:
            vrt = WarpedVRT(src, crs=self.crs)
            src.close()
            return vrt
        return src

    def _read_tif(self, **kwargs) -> None:
        return self._read(self.path, **kwargs)

    @functools.lru_cache(maxsize=128)
    def _read(self, path: str | Path, **kwargs) -> None:
        with opener(path, file_system=self.file_system) as file:
            with rasterio.open(file) as src:
                self._add_meta_from_src(src)
                out_shape = self._get_shape_from_res(self.res)

                if hasattr(self, "_warped_crs"):
                    src = WarpedVRT(src, crs=self.crs)

                self.values = src.read(
                    out_shape=out_shape,
                    **(self.read_kwargs | kwargs),
                )
                if self._dtype:
                    self = self.astype(self.dtype)
                else:
                    self = self.as_minimum_dtype()

    def _read_with_mask(
        self, mask: Any, masked: bool, boundless: bool, **kwargs
    ) -> None:
        kwargs["mask"] = mask

        def _read(self, src, mask, **kwargs):
            self._add_meta_from_src(src)
            if self.bounds is None:
                self._bounds = to_bbox(mask)

            window = rasterio.windows.from_bounds(
                *to_bbox(mask), transform=self.transform
            )

            out_shape = get_shape_from_bounds(mask, self.res)

            kwargs = (
                {"window": window, "boundless": boundless} | self.read_kwargs | kwargs
            )

            if hasattr(self, "_warped_crs"):
                src = WarpedVRT(src, crs=self.crs)

            self.values = src.read(out_shape=out_shape, **kwargs)

            if not masked:
                try:
                    self.values[self.values.mask] = self.nodata
                    self.values = self.values.data
                except AttributeError:
                    pass
                    # self.values = np.ma.masked_array(self.values, mask=mask)
                    # self.values[self.values.mask] = self.nodata
                    # self.values = self.values.data

            if boundless:
                self._bounds = src.window_bounds(window=window)
            else:
                intersected = to_shapely(self.bounds).intersection(to_shapely(mask))
                if intersected.is_empty:
                    self._bounds = None
                else:
                    self._bounds = intersected.bounds

            if not np.size(self.values):
                return

            if self._dtype:
                self = self.astype(self._dtype)
            else:
                self = self.as_minimum_dtype()

        if self.values is not None:
            with memfile_from_array(self.values, **self.profile) as src:
                _read(self, src, **kwargs)
        else:
            with opener(self.path, file_system=self.file_system) as file:
                with rasterio.open(file, **self.profile) as src:
                    _read(self, src, **kwargs)

    def _check_for_array(self, text=""):
        if self.values is None:
            raise ValueError("Arrays are not loaded. " + text)


def get_transform_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, shape: tuple[float, ...]
) -> Affine:
    minx, miny, maxx, maxy = to_bbox(obj)
    if len(shape) == 2:
        width, height = shape
    elif len(shape) == 3:
        _, width, height = shape
    else:
        raise ValueError
    return rasterio.transform.from_bounds(minx, miny, maxx, maxy, width, height)


def get_shape_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, res: int
) -> tuple[int, int]:
    resx, resy = (res, res) if isinstance(res, numbers.Number) else res

    minx, miny, maxx, maxy = to_bbox(obj)
    diffx = maxx - minx
    diffy = maxy - miny
    width = int(diffx / resx)
    heigth = int(diffy / resy)
    return heigth, width



import functools
import glob
import itertools
import numbers
import os
import random
import re
import warnings
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence
from copy import deepcopy
from json import loads
from pathlib import Path
from typing import Any
from typing import ClassVar

import joblib
import numpy as np
import pandas as pd
import pyproj
import rasterio
from affine import Affine
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from rasterio import features
from rasterio.enums import MergeAlg
from rtree.index import Index
from rtree.index import Property
from shapely import Geometry
from shapely import box
from shapely import unary_union
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import shape

try:
    import dapla as dp
    from dapla.gcs import GCSFileSystem
except ImportError:

    class GCSFileSystem:
        """Placeholder."""


try:
    from rioxarray.exceptions import NoDataInBounds
    from rioxarray.merge import merge_arrays
    from rioxarray.rioxarray import _generate_spatial_coords
except ImportError:
    pass
try:
    import xarray as xr
    from xarray import DataArray
except ImportError:

    class DataArray:
        """Placeholder."""


try:
    import torch
except ImportError:
    pass

try:
    from gcsfs.core import GCSFile
except ImportError:

    class GCSFile:
        """Placeholder."""


try:
    from torchgeo.datasets.utils import disambiguate_timestamp
except ImportError:

    class torch:
        """Placeholder."""

        class Tensor:
            """Placeholder to reference torch.Tensor."""


try:
    from torchgeo.datasets.utils import BoundingBox
except ImportError:

    class BoundingBox:
        """Placeholder."""

        def __init__(self, *args, **kwargs) -> None:
            """Placeholder."""
            raise ImportError("missing optional dependency 'torchgeo'")


from ..geopandas_tools.bounds import get_total_bounds
from ..geopandas_tools.bounds import to_bbox
from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.conversion import to_shapely
from ..geopandas_tools.general import get_common_crs
from ..helpers import get_all_files
from ..helpers import get_numpy_func
from ..io._is_dapla import is_dapla
from ..io.opener import opener
from . import sentinel_config as config
from .base import get_index_mapper
from .indices import ndvi
from .zonal import _aggregate
from .zonal import _make_geometry_iterrows
from .zonal import _no_overlap_df
from .zonal import _prepare_zonal
from .zonal import _zonal_post

if is_dapla():

    def ls_func(*args, **kwargs) -> list[str]:
        return dp.FileClient.get_gcs_file_system().ls(*args, **kwargs)

    def glob_func(*args, **kwargs) -> list[str]:
        return dp.FileClient.get_gcs_file_system().glob(*args, **kwargs)

    def open_func(*args, **kwargs) -> GCSFile:
        return dp.FileClient.get_gcs_file_system().open(*args, **kwargs)

else:
    ls_func = functools.partial(get_all_files, recursive=False)
    open_func = open
    glob_func = glob.glob

TORCHGEO_RETURN_TYPE = dict[str, torch.Tensor | pyproj.CRS | BoundingBox]
FILENAME_COL_SUFFIX = "_filename"
DEFAULT_FILENAME_REGEX = r".*\.(?:tif|tiff|jp2)$"


class ImageCollectionGroupBy:
    """Iterator and merger class returned from groupby.

    Can be iterated through like pandas.DataFrameGroupBy.
    Or use the methods merge_by_band or merge.
    """

    def __init__(
        self,
        data: Iterable[tuple[Any], "ImageCollection"],
        by: list[str],
        collection: "ImageCollection",
    ) -> None:
        """Initialiser.

        Args:
            data: Iterable of group values and ImageCollection groups.
            by: list of group attributes.
            collection: ImageCollection instance. Used to pass attributes.
        """
        self.data = list(data)
        self.by = by
        self.collection = collection

    def merge_by_band(
        self, bounds=None, method="median", as_int: bool = True, indexes=None, **kwargs
    ) -> "ImageCollection":
        """Merge each group into separate Bands per band_id, returned as an ImageCollection."""
        images = self._run_func_for_collection_groups(
            _merge_by_band,
            method=method,
            bounds=bounds,
            as_int=as_int,
            indexes=indexes,
            **kwargs,
        )
        print("hihihih")
        for img, (group_values, _) in zip(images, self.data, strict=True):
            for attr, group_value in zip(self.by, group_values, strict=True):
                print(attr, group_value)
                try:
                    setattr(img, attr, group_value)
                except AttributeError:
                    setattr(img, f"_{attr}", group_value)

        collection = ImageCollection(
            images,
            level=self.collection.level,
            **self.collection._common_init_kwargs,
        )
        collection._merged = True
        return collection

    def merge(
        self, bounds=None, method="median", as_int: bool = True, indexes=None, **kwargs
    ) -> "Image":
        """Merge each group into a single Band, returned as combined Image."""
        bands: list[Band] = self._run_func_for_collection_groups(
            _merge,
            method=method,
            bounds=bounds,
            as_int=as_int,
            indexes=indexes,
            **kwargs,
        )
        for band, (group_values, _) in zip(bands, self.data, strict=True):
            for attr, group_value in zip(self.by, group_values, strict=True):
                try:
                    setattr(band, attr, group_value)
                except AttributeError:
                    if hasattr(band, f"_{attr}"):
                        setattr(band, f"_{attr}", group_value)

        if "band_id" in self.by:
            for band in bands:
                assert band.band_id is not None

        image = Image(
            bands,
            **self.collection._common_init_kwargs,
        )
        image._merged = True
        return image

    def _run_func_for_collection_groups(self, func: Callable, **kwargs) -> list[Any]:
        if self.collection.processes == 1:
            return [func(group, **kwargs) for _, group in self]
        processes = min(self.collection.processes, len(self))

        if processes == 0:
            return []

        with joblib.Parallel(n_jobs=processes, backend="threading") as parallel:
            return parallel(joblib.delayed(func)(group, **kwargs) for _, group in self)

    def __iter__(self) -> Iterator[tuple[tuple[Any, ...], "ImageCollection"]]:
        """Iterate over the group values and the ImageCollection groups themselves."""
        return iter(self.data)

    def __len__(self) -> int:
        """Number of ImageCollection groups."""
        return len(self.data)

    def __repr__(self) -> str:
        """String representation."""
        return f"{self.__class__.__name__}({len(self)})"


class _ImageBase:
    image_regexes: ClassVar[str | None] = None
    filename_regexes: ClassVar[str | tuple[str]] = (DEFAULT_FILENAME_REGEX,)
    date_format: ClassVar[str] = "%Y%m%d"  # T%H%M%S"

    def __init__(self) -> None:

        self._merged = False
        self._from_array = False
        self._from_gdf = False

        if self.filename_regexes:
            if isinstance(self.filename_regexes, str):
                self.filename_regexes = (self.filename_regexes,)
            self.filename_patterns = [
                re.compile(regexes, flags=re.VERBOSE)
                for regexes in self.filename_regexes
            ]
        else:
            self.filename_patterns = None

        if self.image_regexes:
            if isinstance(self.image_regexes, str):
                self.image_regexes = (self.image_regexes,)
            self.image_patterns = [
                re.compile(regexes, flags=re.VERBOSE) for regexes in self.image_regexes
            ]
        else:
            self.image_patterns = None

    @property
    def _common_init_kwargs(self) -> dict:
        return {
            "file_system": self.file_system,
            "processes": self.processes,
            "res": self.res,
            "_mask": self._mask,
        }

    @property
    def path(self) -> str:
        try:
            return self._path
        except AttributeError as e:
            raise PathlessImageError(self.__class__.__name__) from e

    @property
    def res(self) -> int:
        """Pixel resolution."""
        return self._res

    @property
    def centroid(self) -> Point:
        return self.unary_union.centroid

    def _name_regex_searcher(
        self, group: str, patterns: tuple[re.Pattern]
    ) -> str | None:
        if not patterns or not any(pat.groups for pat in patterns):
            return None
        for pat in patterns:
            try:
                return re.match(pat, self.name).group(group)
            except (AttributeError, TypeError, IndexError):
                pass
        raise ValueError(
            f"Couldn't find group '{group}' in name {self.name} with regex patterns {patterns}"
        )

    def _create_metadata_df(self, file_paths: list[str]) -> None:
        df = pd.DataFrame({"file_path": file_paths})

        df["filename"] = df["file_path"].apply(lambda x: _fix_path(Path(x).name))
        if not self.single_banded:
            df["image_path"] = df["file_path"].apply(
                lambda x: _fix_path(str(Path(x).parent))
            )
        else:
            df["image_path"] = df["file_path"]

        if not len(df):
            return df

        if self.filename_patterns:
            df, match_cols_filename = _get_regexes_matches_for_df(
                df, "filename", self.filename_patterns, suffix=FILENAME_COL_SUFFIX
            )

            if not len(df):
                return df

            self._match_cols_filename = match_cols_filename
            grouped = (
                df.drop(columns=match_cols_filename, errors="ignore")
                .drop_duplicates("image_path")
                .set_index("image_path")
            )
            for col in ["file_path", "filename", *match_cols_filename]:
                if col in df:
                    grouped[col] = df.groupby("image_path")[col].apply(tuple)

            grouped = grouped.reset_index()
        else:
            df["file_path"] = df.groupby("image_path")["file_path"].apply(tuple)
            df["filename"] = df.groupby("image_path")["filename"].apply(tuple)
            grouped = df.drop_duplicates("image_path")

        grouped["imagename"] = grouped["image_path"].apply(
            lambda x: _fix_path(Path(x).name)
        )

        if self.image_patterns and len(grouped):
            grouped, _ = _get_regexes_matches_for_df(
                grouped, "imagename", self.image_patterns, suffix=""
            )

        if "date" in grouped:
            return grouped.sort_values("date")
        else:
            return grouped

    def copy(self) -> "_ImageBase":
        """Copy the instance and its attributes."""
        copied = deepcopy(self)
        for key, value in copied.__dict__.items():
            try:
                setattr(copied, key, value.copy())
            except AttributeError:
                setattr(copied, key, deepcopy(value))
            except TypeError:
                continue
        return copied


class _ImageBandBase(_ImageBase):
    def intersects(self, other: GeoDataFrame | GeoSeries | Geometry) -> bool:
        if hasattr(other, "crs") and not pyproj.CRS(self.crs).equals(
            pyproj.CRS(other.crs)
        ):
            raise ValueError(f"crs mismatch: {self.crs} and {other.crs}")
        return self.unary_union.intersects(to_shapely(other))

    @property
    def year(self) -> str:
        return self.date[:4]

    @property
    def month(self) -> str:
        return "".join(self.date.split("-"))[:6]

    @property
    def yyyymmd(self) -> str:
        return "".join(self.date.split("-"))[:7]

    @property
    def name(self) -> str | None:
        if hasattr(self, "_name") and self._name is not None:
            return self._name
        try:
            return Path(self.path).name
        except (ValueError, AttributeError):
            return None

    @name.setter
    def name(self, value) -> None:
        self._name = value

    @property
    def stem(self) -> str | None:
        try:
            return Path(self.path).stem
        except (AttributeError, ValueError):
            return None

    @property
    def level(self) -> str:
        return self._name_regex_searcher("level", self.image_patterns)

    @property
    def mint(self) -> float:
        return disambiguate_timestamp(self.date, self.date_format)[0]

    @property
    def maxt(self) -> float:
        return disambiguate_timestamp(self.date, self.date_format)[1]

    @property
    def unary_union(self) -> Polygon:
        return box(*self.bounds)

    @property
    def bbox(self) -> BoundingBox:
        bounds = GeoSeries([self.unary_union]).bounds
        return BoundingBox(
            minx=bounds.minx[0],
            miny=bounds.miny[0],
            maxx=bounds.maxx[0],
            maxy=bounds.maxy[0],
            mint=self.mint,
            maxt=self.maxt,
        )


class Band(_ImageBandBase):
    """Band holding a single 2 dimensional array representing an image band."""

    cmap: ClassVar[str | None] = None

    def __init__(
        self,
        data: str | np.ndarray,
        res: int | None,
        crs: Any | None = None,
        bounds: tuple[float, float, float, float] | None = None,
        cmap: str | None = None,
        name: str | None = None,
        file_system: GCSFileSystem | None = None,
        band_id: str | None = None,
        processes: int = 1,
        _mask: GeoDataFrame | GeoSeries | Geometry | tuple[float] | None = None,
        **kwargs,
    ) -> None:
        """Band initialiser."""
        if isinstance(data, (GeoDataFrame | GeoSeries)):
            if res is None:
                raise ValueError("Must specify res when data is vector geometries.")
            bounds = to_bbox(bounds) if bounds is not None else data.total_bounds
            crs = crs if crs else data.crs
            data: np.ndarray = _arr_from_gdf(data, res=res, **kwargs)
            self._from_gdf = True

        if isinstance(data, np.ndarray):
            self._values = data
            if bounds is None:
                raise ValueError("Must specify bounds when data is an array.")
            self._bounds = to_bbox(bounds)
            self._crs = crs
            self.transform = _get_transform_from_bounds(
                self.bounds, shape=self.values.shape
            )
            self._from_array = True

        elif not isinstance(data, (str | Path | os.PathLike)):
            raise TypeError(
                "'data' must be string, Path-like or numpy.ndarray. "
                f"Got {type(data)}"
            )
        else:
            self._path = str(data)

        self._res = res
        if cmap is not None:
            self.cmap = cmap
        self.file_system = file_system
        self._mask = _mask
        self._name = name
        self._band_id = band_id
        self.processes = processes

        if self.filename_regexes:
            if isinstance(self.filename_regexes, str):
                self.filename_regexes = [self.filename_regexes]
            self.filename_patterns = [
                re.compile(pat, flags=re.VERBOSE) for pat in self.filename_regexes
            ]
        else:
            self.filename_patterns = None

    def __lt__(self, other: "Band") -> bool:
        """Makes Bands sortable by band_id."""
        return self.band_id < other.band_id

    @property
    def values(self) -> np.ndarray:
        """The numpy array, if loaded."""
        try:
            return self._values
        except AttributeError as e:
            raise ValueError("array is not loaded.") from e

    @values.setter
    def values(self, new_val):
        if isinstance(new_val, np.ndarray):
            raise TypeError(f"{self.__class__.__name__} 'values' must be np.ndarray.")
        self._values = new_val

    @property
    def band_id(self) -> str:
        """Band id."""
        if self._band_id is not None:
            return self._band_id
        return self._name_regex_searcher("band", self.filename_patterns)

    @property
    def height(self) -> int:
        """Pixel heigth of the image band."""
        i = 1 if len(self.values.shape) == 3 else 0
        return self.values.shape[i]

    @property
    def width(self) -> int:
        """Pixel width of the image band."""
        i = 2 if len(self.values.shape) == 3 else 1
        return self.values.shape[i]

    @property
    def tile(self) -> str:
        """Tile name from filename_regex."""
        if hasattr(self, "_tile") and self._tile:
            return self._tile
        return self._name_regex_searcher("tile", self.filename_patterns)

    @property
    def date(self) -> str:
        """Tile name from filename_regex."""
        if hasattr(self, "_date") and self._date:
            return self._date

        return self._name_regex_searcher("date", self.filename_patterns)

    @property
    def crs(self) -> str | None:
        """Coordinate reference system."""
        try:
            return self._crs
        except AttributeError:
            with opener(self.path, file_system=self.file_system) as file:
                with rasterio.open(file) as src:
                    self._bounds = to_bbox(src.bounds)
                    self._crs = src.crs
                    return self._crs

    @property
    def bounds(self) -> tuple[int, int, int, int] | None:
        """Bounds as tuple (minx, miny, maxx, maxy)."""
        try:
            return self._bounds
        except AttributeError:
            with opener(self.path, file_system=self.file_system) as file:
                with rasterio.open(file) as src:
                    self._bounds = to_bbox(src.bounds)
                    self._crs = src.crs
                    return self._bounds
        except TypeError:
            return None

    def get_n_largest(
        self, n: int, precision: float = 0.000001, column: str = "value"
    ) -> GeoDataFrame:
        """Get the largest values of the array as polygons in a GeoDataFrame."""
        copied = self.copy()
        value_must_be_at_least = np.sort(np.ravel(copied.values))[-n] - (precision or 0)
        copied._values = np.where(copied.values >= value_must_be_at_least, 1, 0)
        df = copied.to_gdf(column).loc[lambda x: x[column] == 1]
        df[column] = f"largest_{n}"
        return df

    def get_n_smallest(
        self, n: int, precision: float = 0.000001, column: str = "value"
    ) -> GeoDataFrame:
        """Get the lowest values of the array as polygons in a GeoDataFrame."""
        copied = self.copy()
        value_must_be_at_least = np.sort(np.ravel(copied.values))[n] - (precision or 0)
        copied._values = np.where(copied.values <= value_must_be_at_least, 1, 0)
        df = copied.to_gdf(column).loc[lambda x: x[column] == 1]
        df[column] = f"smallest_{n}"
        return df

    def load(self, bounds=None, indexes=None, **kwargs) -> "Band":
        """Load and potentially clip the array.

        The array is stored in the 'values' property.
        """
        bounds = to_bbox(bounds) if bounds is not None else self._mask

        try:
            assert isinstance(self.values, np.ndarray)
            has_array = True
        except (ValueError, AssertionError):
            has_array = False

        if has_array:
            if bounds is None:
                return self
            # bounds_shapely = to_shapely(bounds)
            # if not bounds_shapely.intersects(self.)
            # bounds_arr = GeoSeries([bounds_shapely]).values
            bounds_arr = GeoSeries([to_shapely(bounds)]).values
            try:
                self._values = (
                    to_xarray(
                        self.values,
                        transform=self.transform,
                        crs=self.crs,
                        name=self.name,
                    )
                    .rio.clip(bounds_arr, crs=self.crs)
                    .to_numpy()
                )
            except NoDataInBounds:
                self._values = np.array([])
            self._bounds = bounds
            return self

        with opener(self.path, file_system=self.file_system) as f:
            with rasterio.open(f) as src:
                self._res = int(src.res[0]) if not self.res else self.res
                # if bounds is None:
                #     out_shape = _get_shape_from_res(to_bbox(src.bounds), self.res, indexes)
                #     self.transform = src.transform
                #     arr = src.load(indexes=indexes, out_shape=out_shape, **kwargs)
                #     # if isinstance(indexes, int) and len(arr.shape) == 3:
                #     #     return arr[0]
                #     return arr
                # else:
                #     window = rasterio.windows.from_bounds(
                #         *bounds, transform=src.transform
                #     )
                #     out_shape = _get_shape_from_bounds(bounds, self.res)

                #     arr = src.read(
                #         indexes=indexes,
                #         out_shape=out_shape,
                #         window=window,
                #         boundless=boundless,
                #         **kwargs,
                #     )
                #     if isinstance(indexes, int):
                #         # arr = arr[0]
                #         height, width = arr.shape
                #     else:
                #         height, width = arr.shape[1:]

                #     self.transform = rasterio.transform.from_bounds(
                #         *bounds, width, height
                #     )
                #     if bounds is not None:
                #         self._bounds = bounds
                #     return arr

                if indexes is None and len(src.indexes) == 1:
                    indexes = 1

                if isinstance(indexes, int):
                    _indexes = (indexes,)
                else:
                    _indexes = indexes

                arr, transform = rasterio.merge.merge(
                    [src],
                    res=self.res,
                    indexes=_indexes,
                    bounds=bounds,
                    **kwargs,
                )
                self.transform = transform
                if bounds is not None:
                    self._bounds = bounds

                if isinstance(indexes, int):
                    arr = arr[0]

                self._values = arr
        return self

    def write(self, path: str | Path, **kwargs) -> None:
        """Write the array as an image file."""
        if not hasattr(self, "_values"):
            raise ValueError(
                "Can only write image band from Band constructed from array."
            )

        if self.crs is None:
            raise ValueError("Cannot write None crs to image.")

        profile = {
            # "driver": self.driver,
            # "compress": self.compress,
            # "dtype": self.dtype,
            "crs": self.crs,
            "transform": self.transform,
            # "nodata": self.nodata,
            # "count": self.count,
            # "height": self.height,
            # "width": self.width,
            # "indexes": self.indexes,
        } | kwargs

        with opener(path, "w", file_system=self.file_system) as f:
            with rasterio.open(f, **profile) as dst:
                # bounds = to_bbox(self._mask) if self._mask is not None else dst.bounds

                # res = dst.res if not self.res else self.res

                if len(self.values.shape) == 2:
                    return dst.write(self.values, indexes=1)

                for i in range(self.values.shape[0]):
                    dst.write(self.values[i], indexes=i + 1)

        self._path = str(path)

    def sample(self, size: int = 1000, mask: Any = None, **kwargs) -> "Image":
        """Take a random spatial sample area of the Band."""
        copied = self.copy()
        if mask is not None:
            point = GeoSeries([copied.unary_union]).clip(mask).sample_points(1)
        else:
            point = GeoSeries([copied.unary_union]).sample_points(1)
        buffered = point.buffer(size / 2).clip(copied.unary_union)
        copied = copied.load(bounds=buffered.total_bounds, **kwargs)
        return copied

    def get_gradient(self, degrees: bool = False, copy: bool = True) -> "Band":
        """Get the slope of an elevation band.

        Calculates the absolute slope between the grid cells
        based on the image resolution.

        For multi-band images, the calculation is done for each band.

        Args:
            band: band instance.
            degrees: If False (default), the returned values will be in ratios,
                where a value of 1 means 1 meter up per 1 meter forward. If True,
                the values will be in degrees from 0 to 90.
            copy: Whether to copy or overwrite the original Raster.
                Defaults to True.

        Returns:
            The class instance with new array values, or a copy if copy is True.

        Examples:
        ---------
        Making an array where the gradient to the center is always 10.

        >>> import sgis as sg
        >>> import numpy as np
        >>> arr = np.array(
        ...         [
        ...             [100, 100, 100, 100, 100],
        ...             [100, 110, 110, 110, 100],
        ...             [100, 110, 120, 110, 100],
        ...             [100, 110, 110, 110, 100],
        ...             [100, 100, 100, 100, 100],
        ...         ]
        ...     )

        Now let's create a Raster from this array with a resolution of 10.

        >>> band = sg.Band(arr, crs=None, bounds=(0, 0, 50, 50), res=10)

        The gradient will be 1 (1 meter up for every meter forward).
        The calculation is by default done in place to save memory.

        >>> band.gradient()
        >>> band.values
        array([[0., 1., 1., 1., 0.],
            [1., 1., 1., 1., 1.],
            [1., 1., 0., 1., 1.],
            [1., 1., 1., 1., 1.],
            [0., 1., 1., 1., 0.]])
        """
        copied = self.copy() if copy else self
        copied._values = _get_gradient(copied, degrees=degrees, copy=copy)
        return copied

    def zonal(
        self,
        polygons: GeoDataFrame,
        aggfunc: str | Callable | list[Callable | str],
        array_func: Callable | None = None,
        dropna: bool = True,
    ) -> GeoDataFrame:
        """Calculate zonal statistics in polygons.

        Args:
            polygons: A GeoDataFrame of polygon geometries.
            aggfunc: Function(s) of which to aggregate the values
                within each polygon.
            array_func: Optional calculation of the raster
                array before calculating the zonal statistics.
            dropna: If True (default), polygons with all missing
                values will be removed.

        Returns:
            A GeoDataFrame with aggregated values per polygon.
        """
        idx_mapper, idx_name = get_index_mapper(polygons)
        polygons, aggfunc, func_names = _prepare_zonal(polygons, aggfunc)
        poly_iter = _make_geometry_iterrows(polygons)

        kwargs = {
            "band": self,
            "aggfunc": aggfunc,
            "array_func": array_func,
            "func_names": func_names,
        }

        if self.processes == 1:
            aggregated = [_zonal_one_pair(i, poly, **kwargs) for i, poly in poly_iter]
        else:
            with joblib.Parallel(n_jobs=self.processes, backend="loky") as parallel:
                aggregated = parallel(
                    joblib.delayed(_zonal_one_pair)(i, poly, **kwargs)
                    for i, poly in poly_iter
                )

        return _zonal_post(
            aggregated,
            polygons=polygons,
            idx_mapper=idx_mapper,
            idx_name=idx_name,
            dropna=dropna,
        )

    def to_gdf(self, column: str = "value") -> GeoDataFrame:
        """Create a GeoDataFrame from the image Band.

        Args:
            column: Name of resulting column that holds the raster values.

        Returns:
            A GeoDataFrame with a geometry column and array values.
        """
        if not hasattr(self, "_values"):
            raise ValueError("Array is not loaded.")

        if self.values.shape[0] == 0:
            return GeoDataFrame({"geometry": []}, crs=self.crs)

        return GeoDataFrame(
            pd.DataFrame(
                _array_to_geojson(
                    self.values, self.transform, processes=self.processes
                ),
                columns=[column, "geometry"],
            ),
            geometry="geometry",
            crs=self.crs,
        )

    def to_xarray(self) -> DataArray:
        """Convert the raster to  an xarray.DataArray."""
        name = self.name or self.__class__.__name__.lower()
        coords = _generate_spatial_coords(self.transform, self.width, self.height)
        if len(self.values.shape) == 2:
            dims = ["y", "x"]
        elif len(self.values.shape) == 3:
            dims = ["band", "y", "x"]
        else:
            raise ValueError("Array must be 2 or 3 dimensional.")
        return xr.DataArray(
            self.values,
            coords=coords,
            dims=dims,
            name=name,
            attrs={"crs": self.crs},
        )  # .transpose("y", "x")

    def __repr__(self) -> str:
        """String representation."""
        try:
            band_id = f"'{self.band_id}'" if self.band_id else None
        except (ValueError, AttributeError):
            band_id = None
        try:
            path = f"'{self.path}'"
        except (ValueError, AttributeError):
            path = None
        return (
            f"{self.__class__.__name__}(band_id={band_id}, res={self.res}, path={path})"
        )


class NDVIBand(Band):
    """Band for NDVI values."""

    cmap: str = "Greens"


class Image(_ImageBandBase):
    """Image consisting of one or more Bands."""

    cloud_cover_regexes: ClassVar[tuple[str] | None] = None
    band_class: ClassVar[Band] = Band

    def __init__(
        self,
        data: str | Path | Sequence[Band],
        res: int | None = None,
        # crs: Any | None = None,
        single_banded: bool = False,
        file_system: GCSFileSystem | None = None,
        df: pd.DataFrame | None = None,
        all_file_paths: list[str] | None = None,
        _mask: GeoDataFrame | GeoSeries | Geometry | tuple | None = None,
        processes: int = 1,
    ) -> None:
        """Image initialiser."""
        super().__init__()

        self._res = res
        # self._crs = crs
        self.file_system = file_system
        self._mask = _mask
        self.single_banded = single_banded
        self.processes = processes

        if hasattr(data, "__iter__") and all(isinstance(x, Band) for x in data):
            self._bands = list(data)
            self._bounds = get_total_bounds(self._bands)
            self._crs = get_common_crs(self._bands)
            res = list({band.res for band in self._bands})
            if len(res) == 1:
                self._res = res[0]
            else:
                raise ValueError(f"Different resolutions for the bands: {res}")
            return

        if not isinstance(data, (str | Path | os.PathLike)):
            raise TypeError("'data' must be string, Path-like or a sequence of Band.")

        self._path = str(data)

        if df is None:
            if is_dapla():
                file_paths = list(sorted(set(glob_func(self.path + "/**"))))
            else:
                file_paths = list(
                    sorted(
                        set(
                            glob_func(self.path + "/**/**")
                            + glob_func(self.path + "/**/**/**")
                            + glob_func(self.path + "/**/**/**/**")
                            + glob_func(self.path + "/**/**/**/**/**")
                        )
                    )
                )
            df = self._create_metadata_df(file_paths)

        df["image_path"] = df["image_path"].astype(str)

        cols_to_explode = [
            "file_path",
            "filename",
            *[x for x in df if FILENAME_COL_SUFFIX in x],
        ]
        try:
            df = df.explode(cols_to_explode, ignore_index=True)
        except ValueError:
            for col in cols_to_explode:
                df = df.explode(col)
            df = df.loc[lambda x: ~x["filename"].duplicated()].reset_index(drop=True)

        df = df.loc[lambda x: x["image_path"].str.contains(_fix_path(self.path))]

        if self.filename_patterns and any(pat.groups for pat in self.filename_patterns):
            df = df.loc[
                lambda x: (x[f"band{FILENAME_COL_SUFFIX}"].notna())
            ].sort_values(f"band{FILENAME_COL_SUFFIX}")

        if self.cloud_cover_regexes:
            if all_file_paths is None:
                file_paths = ls_func(self.path)
            else:
                file_paths = [path for path in all_file_paths if self.name in path]
            self.cloud_coverage_percentage = float(
                _get_regex_match_from_xml_in_local_dir(
                    file_paths, regexes=self.cloud_cover_regexes
                )
            )
        else:
            self.cloud_coverage_percentage = None

        self._bands = [
            self.band_class(
                path,
                **self._common_init_kwargs,
            )
            for path in (df["file_path"])
        ]

        if self.filename_patterns and any(pat.groups for pat in self.filename_patterns):
            self._bands = list(sorted(self._bands))

    def get_ndvi(self, red_band: str, nir_band: str) -> NDVIBand:
        """Calculate the NDVI for the Image."""
        red = self[red_band].load().values
        nir = self[nir_band].load().values

        arr: np.ndarray = ndvi(red, nir)

        return NDVIBand(
            arr,
            bounds=self.bounds,
            crs=self.crs,
            **self._common_init_kwargs,
        )

    def get_brightness(self, bounds=None, rbg_bands: list[str] | None = None) -> Band:
        """Get a Band with a brightness score of the Image's RBG bands."""
        if rbg_bands is None:
            try:
                r, b, g = self.rbg_bands
            except AttributeError as err:
                raise AttributeError(
                    "Must specify rbg_bands when there is no class variable 'rbd_bands'"
                ) from err
        else:
            r, b, g = rbg_bands

        red = self[r].load(bounds=bounds)
        blue = self[b].load(bounds=bounds)
        green = self[g].load(bounds=bounds)

        brightness = (
            0.299 * red.values + 0.587 * green.values + 0.114 * blue.values
        ).astype(int)

        return Band(
            brightness,
            bounds=red.bounds,
            crs=self.crs,
            **self._common_init_kwargs,
        )

    @property
    def band_ids(self) -> list[str]:
        """The Band ids."""
        return [band.band_id for band in self]

    @property
    def file_paths(self) -> list[str]:
        """The Band file paths."""
        return [band.path for band in self]

    @property
    def bands(self) -> list[Band]:
        """The Image Bands."""
        return self._bands

    @property
    def tile(self) -> str:
        """Tile name from filename_regex."""
        if hasattr(self, "_tile") and self._tile:
            return self._tile
        return self._name_regex_searcher("tile", self.image_patterns)

    @property
    def date(self) -> str:
        """Tile name from filename_regex."""
        if hasattr(self, "_date") and self._date:
            return self._date

        return self._name_regex_searcher("date", self.image_patterns)

    @property
    def crs(self) -> str | None:
        """Coordinate reference system of the Image."""
        try:
            return self._crs
        except AttributeError:
            if not len(self):
                return None
            with opener(self.file_paths[0], file_system=self.file_system) as file:
                with rasterio.open(file) as src:
                    self._bounds = to_bbox(src.bounds)
                    self._crs = src.crs
                    return self._crs

    @property
    def bounds(self) -> tuple[int, int, int, int] | None:
        """Bounds of the Image (minx, miny, maxx, maxy)."""
        try:
            return self._bounds
        except AttributeError:
            if not len(self):
                return None
            with opener(self.file_paths[0], file_system=self.file_system) as file:
                with rasterio.open(file) as src:
                    self._bounds = to_bbox(src.bounds)
                    self._crs = src.crs
                    return self._bounds
        except TypeError:
            return None

    # @property
    # def year(self) -> str:
    #     return self.date[:4]

    # @property
    # def month(self) -> str:
    #     return "".join(self.date.split("-"))[:6]

    # def write(self, image_path: str | Path, file_type: str = "tif", **kwargs) -> None:
    #     _test = kwargs.pop("_test")
    #     suffix = "." + file_type.strip(".")
    #     img_path = Path(image_path)
    #     for band in self:
    #         band_path = (img_path / band.name).with_suffix(suffix)
    #         if _test:
    #             print(f"{self.__class__.__name__}.write: {band_path}")
    #             continue

    #         band.write(band_path, **kwargs)

    # def read(self, bounds=None, **kwargs) -> np.ndarray:
    #     """Return 3 dimensional numpy.ndarray of shape (n bands, width, height)."""
    #     return np.array(
    #         [(band.load(bounds=bounds, **kwargs).values) for band in self.bands]
    #     )

    def to_gdf(self, column: str = "value") -> GeoDataFrame:
        """Convert the array to a GeoDataFrame of grid polygons and values."""
        return pd.concat(
            [band.to_gdf(column=column) for band in self], ignore_index=True
        )

    def sample(
        self, n: int = 1, size: int = 1000, mask: Any = None, **kwargs
    ) -> "Image":
        """Take a random spatial sample of the image."""
        copied = self.copy()
        if mask is not None:
            points = GeoSeries([self.unary_union]).clip(mask).sample_points(n)
        else:
            points = GeoSeries([self.unary_union]).sample_points(n)
        buffered = points.buffer(size / 2).clip(self.unary_union)
        boxes = to_gdf([box(*arr) for arr in buffered.bounds.values], crs=self.crs)
        copied._bands = [band.load(bounds=boxes, **kwargs) for band in copied]
        return copied

    # def get_filepath(self, band: str) -> str:
    #     simple_string_match = [path for path in self.file_paths if str(band) in path]
    #     if len(simple_string_match) == 1:
    #         return simple_string_match[0]

    #     regexes_matches = []
    #     for path in self.file_paths:
    #         for pat in self.filename_patterns:
    #             match_ = re.search(pat, Path(path).name)
    #             if match_ and str(band) == match_.group("band"):
    #                 regexes_matches.append(path)

    #     if len(regexes_matches) == 1:
    #         return regexes_matches[0]

    #     if len(regexes_matches) > 1:
    #         prefix = "Multiple"
    #     elif not regexes_matches:
    #         prefix = "No"

    #     raise KeyError(
    #         f"{prefix} matches for band {band} among paths {[Path(x).name for x in self.file_paths]}"
    #     )

    def __getitem__(
        self, band: str | int | Sequence[str] | Sequence[int]
    ) -> "Band | Image":
        """Get bands by band_id or integer index.

        Returns a Band if a string or int is passed,
        returns an Image if a sequence of strings or integers is passed.
        """
        if isinstance(band, str):
            return self._get_band(band)
        if isinstance(band, int):
            return self.bands[band]  # .copy()

        copied = self.copy()
        try:
            copied._bands = [copied._get_band(x) for x in band]
        except TypeError:
            try:
                copied._bands = [copied.bands[i] for i in band]
            except TypeError as e:
                raise TypeError(
                    f"{self.__class__.__name__} indices should be string, int "
                    f"or sequence of string or int. Got {band}."
                ) from e
        return copied

    def __contains__(self, item: str | Sequence[str]) -> bool:
        """Check if the Image contains a band_id (str) or all band_ids in a sequence."""
        if isinstance(item, str):
            return item in self.band_ids
        return all(x in self.band_ids for x in item)

    def __lt__(self, other: "Image") -> bool:
        """Makes Images sortable by date."""
        return self.date < other.date

    def __iter__(self) -> Iterator[Band]:
        """Iterate over the Bands."""
        return iter(self.bands)

    def __len__(self) -> int:
        """Number of bands in the Image."""
        return len(self.bands)

    def __repr__(self) -> str:
        """String representation."""
        return f"{self.__class__.__name__}(bands={self.bands})"

    def get_cloud_band(self) -> Band:
        """Get a Band where self.cloud_values have value 1 and the rest have value 0."""
        scl = self[self.cloud_band].load()
        scl._values = np.where(np.isin(scl.values, self.cloud_values), 1, 0)
        return scl

    # @property
    # def transform(self) -> Affine | None:
    #     """Get the Affine transform of the image."""
    #     try:
    #         return rasterio.transform.from_bounds(*self.bounds, self.width, self.height)
    #     except (ZeroDivisionError, TypeError):
    #         if not self.width or not self.height:
    #             return None

    # @property
    # def shape(self) -> tuple[int]:
    #     return self._shape

    # @property
    # def height(self) -> int:
    #     i = 1 if len(self.shape) == 3 else 0
    #     return self.shape[i]

    # @property
    # def width(self) -> int:
    #     i = 2 if len(self.shape) == 3 else 1
    #     return self.shape[i]

    # @transform.setter
    # def transform(self, value: Affine) -> None:
    #     self._bounds = rasterio.transform.array_bounds(self.height, self.width, value)

    def _get_band(self, band: str) -> Band:
        if not isinstance(band, str):
            raise TypeError(f"band must be string. Got {type(band)}")

        bands = [x for x in self.bands if x.band_id == band]
        if len(bands) == 1:
            return bands[0]
        if len(bands) > 1:
            raise ValueError(
                f"Multiple matches for band_id {band} among {[x for x in self]}"
            )

        try:
            more_bands = [x for x in self.bands if x.path == band]
        except PathlessImageError:
            more_bands = bands

        if len(more_bands) == 1:
            return more_bands[0]

        if len(bands) > 1:
            prefix = "Multiple"
        elif not bands:
            prefix = "No"

        raise KeyError(
            f"{prefix} matches for band {band} among paths {[Path(band.path).name for band in self.bands]}"
        )


class ImageCollection(_ImageBase):
    """Collection of Images.

    Loops though Images.
    """

    image_class: ClassVar[Image] = Image
    band_class: ClassVar[Band] = Band

    def __init__(
        self,
        data: str | Path | Sequence[Image],
        res: int,
        level: str | None,
        single_banded: bool = False,
        processes: int = 1,
        file_system: GCSFileSystem | None = None,
        df: pd.DataFrame | None = None,
        _mask: Any | None = None,
    ) -> None:
        """Initialiser."""
        super().__init__()

        self.level = level
        self.processes = processes
        self.file_system = file_system
        self._res = res
        self._mask = _mask
        self._band_ids = None
        self.single_banded = single_banded

        if hasattr(data, "__iter__") and all(isinstance(x, Image) for x in data):
            self._path = None
            self.images = data
            return

        if not isinstance(data, (str | Path | os.PathLike)):
            raise TypeError("'data' must be string, Path-like or a sequence of Image.")

        self._path = str(data)

        if is_dapla():
            self._all_filepaths = list(sorted(set(glob_func(self.path + "/**"))))
        else:
            self._all_filepaths = list(
                sorted(
                    set(
                        glob_func(self.path + "/**/**")
                        + glob_func(self.path + "/**/**/**")
                        + glob_func(self.path + "/**/**/**/**")
                        + glob_func(self.path + "/**/**/**/**/**")
                    )
                )
            )

        if self.level:
            self._all_filepaths = [
                path for path in self._all_filepaths if self.level in path
            ]

        if df is not None:
            self._df = df
        else:
            self._df = self._create_metadata_df(self._all_filepaths)

    def groupby(self, by: str | list[str], **kwargs) -> ImageCollectionGroupBy:
        """Group the Collection by Image or Band attribute(s)."""
        df = pd.DataFrame(
            [(i, img) for i, img in enumerate(self) for _ in img],
            columns=["_image_idx", "_image_instance"],
        )

        if isinstance(by, str):
            by = [by]

        for attr in by:
            if attr == "bounds":
                # need integers to check equality when grouping
                df[attr] = [
                    tuple(int(x) for x in band.bounds) for img in self for band in img
                ]
                continue

            try:
                df[attr] = [getattr(band, attr) for img in self for band in img]
            except AttributeError:
                df[attr] = [getattr(img, attr) for img in self for _ in img]

        with joblib.Parallel(n_jobs=self.processes, backend="loky") as parallel:
            return ImageCollectionGroupBy(
                sorted(
                    parallel(
                        joblib.delayed(_copy_and_add_df_parallel)(i, group, self)
                        for i, group in df.groupby(by, **kwargs)
                    )
                ),
                by=by,
                collection=self,
            )

    def explode(self, copy: bool = True) -> "ImageCollection":
        """Make all Images single-banded."""
        copied = self.copy() if copy else self
        copied.images = [
            self.image_class(
                [band],
                single_banded=True,
                **self._common_init_kwargs,
                df=self._df,
                all_file_paths=self._all_filepaths,
            )
            for img in self
            for band in img
        ]
        return copied

    def merge(
        self, bounds=None, method="median", as_int: bool = True, indexes=None, **kwargs
    ) -> Band:
        """Merge all areas and all bands to a single Band."""
        bounds = to_bbox(bounds) if bounds is not None else self._mask
        crs = self.crs

        if indexes is None:
            indexes = 1

        if isinstance(indexes, int):
            _indexes = (indexes,)
        else:
            _indexes = indexes

        if method == "mean":
            _method = "sum"
        else:
            _method = method

        if method not in list(rasterio.merge.MERGE_METHODS) + ["mean"]:
            arr = self._merge_with_numpy_func(
                method=method,
                bounds=bounds,
                as_int=as_int,
            )
        else:
            datasets = [_open_raster(path) for path in self.file_paths]
            arr, _ = rasterio.merge.merge(
                datasets,
                res=self.res,
                bounds=bounds,
                indexes=_indexes,
                method=_method,
                **kwargs,
            )

        if isinstance(indexes, int) and len(arr.shape) == 3 and arr.shape[0] == 1:
            arr = arr[0]

        if method == "mean":
            if as_int:
                arr = arr // len(datasets)
            else:
                arr = arr / len(datasets)

        if bounds is None:
            bounds = self.bounds

        # return self.band_class(
        band = Band(
            arr,
            bounds=bounds,
            crs=crs,
            **self._common_init_kwargs,
        )

        band._merged = True
        return band

    def merge_by_band(
        self,
        bounds: tuple | Geometry | GeoDataFrame | GeoSeries | None = None,
        method: str = "median",
        as_int: bool = True,
        indexes: int | tuple[int] | None = None,
        **kwargs,
    ) -> Image:
        """Merge all areas to a single tile, one band per band_id."""
        bounds = to_bbox(bounds) if bounds is not None else self._mask
        bounds = self.bounds if bounds is None else bounds
        out_bounds = self.bounds if bounds is None else bounds
        crs = self.crs

        if indexes is None:
            indexes = 1

        if isinstance(indexes, int):
            _indexes = (indexes,)
        else:
            _indexes = indexes

        if method == "mean":
            _method = "sum"
        else:
            _method = method

        arrs = []
        bands: list[Band] = []
        for (band_id,), band_collection in self.groupby("band_id"):
            if method not in list(rasterio.merge.MERGE_METHODS) + ["mean"]:
                arr = band_collection._merge_with_numpy_func(
                    method=method,
                    bounds=bounds,
                    as_int=as_int,
                )
            else:
                datasets = [_open_raster(path) for path in band_collection.file_paths]
                arr, _ = rasterio.merge.merge(
                    datasets,
                    res=self.res,
                    bounds=bounds,
                    indexes=_indexes,
                    method=_method,
                    **kwargs,
                )
                if isinstance(indexes, int):
                    arr = arr[0]
                if method == "mean":
                    if as_int:
                        arr = arr // len(datasets)
                    else:
                        arr = arr / len(datasets)

            arrs.append(arr)
            bands.append(
                self.band_class(
                    arr,
                    bounds=out_bounds,
                    crs=crs,
                    band_id=band_id,
                    **self._common_init_kwargs,
                )
            )

        # return self.image_class(
        image = Image(
            bands,
            **self._common_init_kwargs,
        )

        image._merged = True
        return image

    def _merge_with_numpy_func(
        self,
        method: str | Callable,
        bounds=None,
        as_int: bool = True,
        indexes=None,
        **kwargs,
    ) -> np.ndarray:
        arrs = []
        numpy_func = get_numpy_func(method) if not callable(method) else method
        for (_bounds,), collection in self.groupby("bounds"):
            arr = np.array(
                [
                    band.load(bounds=bounds, indexes=indexes, **kwargs).values
                    for img in collection
                    for band in img
                ]
            )
            arr = numpy_func(arr, axis=0)
            if as_int:
                arr = arr.astype(int)
                min_dtype = rasterio.dtypes.get_minimum_dtype(arr)
                arr = arr.astype(min_dtype)

            if len(arr.shape) == 2:
                height, width = arr.shape
            elif len(arr.shape) == 3:
                height, width = arr.shape[1:]
            else:
                raise ValueError(arr.shape)

            transform = rasterio.transform.from_bounds(*_bounds, width, height)
            coords = _generate_spatial_coords(transform, width, height)

            arrs.append(
                xr.DataArray(
                    arr,
                    coords=coords,
                    dims=["y", "x"],
                    name=str(_bounds),
                    attrs={"crs": self.crs},
                )
            )

        if bounds is None:
            bounds = self.bounds

        merged = merge_arrays(arrs, bounds=bounds, res=self.res)

        return merged.to_numpy()

    # def write(self, root: str | Path, file_type: str = "tif", **kwargs) -> None:
    #     _test = kwargs.pop("_test")
    #     suffix = "." + file_type.strip(".")
    #     for img in self:
    #         img_path = Path(root) / img.name
    #         for band in img:
    #             band_path = (img_path / band.name).with_suffix(suffix)
    #             if _test:
    #                 print(f"{self.__class__.__name__}.write: {band_path}")
    #                 continue
    #             band.write(band_path, **kwargs)

    def load_bands(self, bounds=None, indexes=None, **kwargs) -> "ImageCollection":
        """Load all image Bands with threading."""
        with joblib.Parallel(n_jobs=self.processes, backend="threading") as parallel:
            parallel(
                joblib.delayed(_load_band)(
                    band, bounds=bounds, indexes=indexes, **kwargs
                )
                for img in self
                for band in img
                if bounds is None or img.intersects(bounds)
            )

        return self

    def set_mask(
        self, mask: GeoDataFrame | GeoSeries | Geometry | tuple[float]
    ) -> "ImageCollection":
        """Set the mask to be used to clip the images to."""
        self._mask = to_bbox(mask)
        # only update images when already instansiated
        if hasattr(self, "_images"):
            for img in self._images:
                img._mask = self._mask
                img._bounds = self._mask
                for band in img:
                    band._mask = self._mask
                    band._bounds = self._mask
        return self

    def filter(
        self,
        bands: str | list[str] | None = None,
        date_ranges: (
            tuple[str | None, str | None]
            | tuple[tuple[str | None, str | None], ...]
            | None
        ) = None,
        bounds: GeoDataFrame | GeoSeries | Geometry | tuple[float] | None = None,
        max_cloud_coverage: int | None = None,
        copy: bool = True,
    ) -> "ImageCollection":
        """Filter images and bands in the collection."""
        copied = self.copy() if copy else self

        if isinstance(bounds, BoundingBox):
            date_ranges = (bounds.mint, bounds.maxt)

        if date_ranges:
            copied = copied._filter_dates(date_ranges, copy=False)

        if max_cloud_coverage is not None:
            copied.images = [
                image
                for image in copied.images
                if image.cloud_coverage_percentage < max_cloud_coverage
            ]

        if bounds is not None:
            copied = copied._filter_bounds(bounds, copy=False)

        if bands is not None:
            if isinstance(bands, str):
                bands = [bands]
            bands = set(bands)
            copied._band_ids = bands
            copied.images = [img[bands] for img in copied.images if bands in img]

        return copied

    def _filter_dates(
        self,
        date_ranges: (
            tuple[str | None, str | None] | tuple[tuple[str | None, str | None], ...]
        ),
        copy: bool = True,
    ) -> "ImageCollection":
        if not isinstance(date_ranges, (tuple, list)):
            raise TypeError(
                "date_ranges should be a 2-length tuple of strings or None, "
                "or a tuple of tuples for multiple date ranges"
            )
        if self.image_patterns is None:
            raise ValueError(
                "Cannot set date_ranges when the class's image_regexes attribute is None"
            )

        copied = self.copy() if copy else self

        copied.images = [
            img
            for img in self
            if _date_is_within(
                img.path, date_ranges, copied.image_patterns, copied.date_format
            )
        ]
        return copied

    def _filter_bounds(
        self, other: GeoDataFrame | GeoSeries | Geometry | tuple, copy: bool = True
    ) -> "ImageCollection":
        copied = self.copy() if copy else self

        other = to_shapely(other)

        with joblib.Parallel(n_jobs=copied.processes, backend="threading") as parallel:
            intersects_list: list[bool] = parallel(
                joblib.delayed(_intesects)(image, other) for image in copied
            )
        copied.images = [
            image
            for image, intersects in zip(copied, intersects_list, strict=False)
            if intersects
        ]
        return copied

    def to_gdfs(self, column: str = "value") -> dict[str, GeoDataFrame]:
        """Convert each band in each Image to a GeoDataFrame."""
        out = {}
        i = 0
        for img in self:
            for band in img:
                i += 1
                try:
                    name = band.name
                except AttributeError:
                    name = f"{self.__class__.__name__}({i})"

                if name not in out:
                    out[name] = band.to_gdf(column=column)
                else:
                    out[name] = f"{self.__class__.__name__}({i})"
        return out

    def sample(self, n: int = 1, size: int = 500) -> "ImageCollection":
        """Sample one or more areas of a given size and set this as mask for the images."""
        images = []
        bbox = to_gdf(self.unary_union).geometry.buffer(-size / 2)
        copied = self.copy()
        for _ in range(n):
            mask = to_bbox(bbox.sample_points(1).buffer(size))
            images += copied.filter(bounds=mask).set_mask(mask).images
        copied.images = images
        return copied

    def sample_tiles(self, n: int) -> "ImageCollection":
        """Sample one or more tiles in a copy of the ImageCollection."""
        copied = self.copy()
        sampled_tiles = list({img.tile for img in self})
        random.shuffle(sampled_tiles)
        sampled_tiles = sampled_tiles[:n]

        copied.images = [image for image in self if image.tile in sampled_tiles]
        return copied

    def sample_images(self, n: int) -> "ImageCollection":
        """Sample one or more images in a copy of the ImageCollection."""
        copied = self.copy()
        images = copied.images
        if n > len(images):
            raise ValueError(
                f"n ({n}) is higher than number of images in collection ({len(images)})"
            )
        sample = []
        for _ in range(n):
            random.shuffle(images)
            img = images.pop()
            sample.append(img)

        copied.images = sample

        return copied

    def __or__(self, collection: "ImageCollection") -> "ImageCollection":
        """Concatenate the collection with another collection."""
        return concat_image_collections([self, collection])

    def __iter__(self) -> Iterator[Image]:
        """Iterate over the images."""
        return iter(self.images)

    def __len__(self) -> int:
        """Number of images."""
        return len(self.images)

    def __getitem__(
        self,
        item: int | slice | Sequence[int | bool] | BoundingBox | Sequence[BoundingBox],
    ) -> Image | TORCHGEO_RETURN_TYPE:
        """Select one Image by integer index, or multiple Images by slice, list of int or torchgeo.BoundingBox."""
        if isinstance(item, int):
            return self.images[item]

        if isinstance(item, slice):
            copied = self.copy()
            copied.images = copied.images[item]
            return copied

        if not isinstance(item, BoundingBox) and not (
            isinstance(item, Iterable) and all(isinstance(x, BoundingBox) for x in item)
        ):
            try:
                copied = self.copy()
                if all(isinstance(x, bool) for x in item):
                    copied.images = [
                        img for x, img in zip(item, copied, strict=True) if x
                    ]
                else:
                    copied.images = [copied.images[i] for i in item]
                return copied
            except Exception as e:
                if hasattr(item, "__iter__"):
                    endnote = f" of length {len(item)} with types {set(type(x) for x in item)}"
                raise TypeError(
                    "ImageCollection indices must be int or BoundingBox. "
                    f"Got {type(item)}{endnote}"
                ) from e

        elif isinstance(item, BoundingBox):
            date_ranges: tuple[str] = (item.mint, item.maxt)
            data: torch.Tensor = numpy_to_torch(
                np.array(
                    [
                        band.values
                        for band in self.filter(
                            bounds=item, date_ranges=date_ranges
                        ).merge_by_band(bounds=item)
                    ]
                )
            )
        else:
            bboxes: list[Polygon] = [to_bbox(x) for x in item]
            date_ranges: list[list[str, str]] = [(x.mint, x.maxt) for x in item]
            data: torch.Tensor = torch.cat(
                [
                    numpy_to_torch(
                        np.array(
                            [
                                band.values
                                for band in self.filter(
                                    bounds=bbox, date_ranges=date_range
                                ).merge_by_band(bounds=bbox)
                            ]
                        )
                    )
                    for bbox, date_range in zip(bboxes, date_ranges, strict=True)
                ]
            )

        crs = get_common_crs(self.images)

        key = "image"  # if self.is_image else "mask"
        sample = {key: data, "crs": crs, "bbox": item}

        return sample

    # def dates_as_float(self) -> list[tuple[float, float]]:
    #     return [disambiguate_timestamp(date, self.date_format) for date in self.dates]

    @property
    def mint(self) -> float:
        """Min timestamp of the images combined."""
        return min(img.mint for img in self)

    @property
    def maxt(self) -> float:
        """Max timestamp of the images combined."""
        return max(img.maxt for img in self)

    @property
    def band_ids(self) -> list[str]:
        """Sorted list of unique band_ids."""
        return list(sorted({band.band_id for img in self for band in img}))

    @property
    def file_paths(self) -> list[str]:
        """Sorted list of all file paths, meaning all band paths."""
        return list(sorted({band.path for img in self for band in img}))

    @property
    def dates(self) -> list[str]:
        """List of image dates."""
        return [img.date for img in self]

    def dates_as_int(self) -> list[int]:
        """List of image dates as 8-length integers."""
        return [int(img.date[:8]) for img in self]

    @property
    def image_paths(self) -> list[str]:
        """List of image paths."""
        return [img.path for img in self]

    @property
    def images(self) -> list["Image"]:
        """List of images in the Collection."""
        try:
            return self._images
        except AttributeError:
            # only fetch images when they are needed
            self._images = _get_images(
                list(self._df["image_path"]),
                all_file_paths=self._all_filepaths,
                df=self._df,
                res=self.res,
                processes=self.processes,
                image_class=self.image_class,
                _mask=self._mask,
            )
            if self.image_regexes:
                self._images = list(sorted(self._images))
            return self._images

    @images.setter
    def images(self, new_value: list["Image"]) -> list["Image"]:
        if self.filename_patterns and any(pat.groups for pat in self.filename_patterns):
            self._images = list(sorted(new_value))
        else:
            self._images = list(new_value)
        if not all(isinstance(x, Image) for x in self._images):
            raise TypeError("images should be a sequence of Image.")

    @property
    def index(self) -> Index:
        """Spatial index that makes torchgeo think this class is a RasterDataset."""
        try:
            if len(self) == len(self._index):
                return self._index
        except AttributeError:
            self._index = Index(interleaved=False, properties=Property(dimension=3))

            for i, img in enumerate(self.images):
                if img.date:
                    try:
                        mint, maxt = disambiguate_timestamp(img.date, self.date_format)
                    except (NameError, TypeError):
                        mint, maxt = 0, 1
                else:
                    mint, maxt = 0, 1
                # important: torchgeo has a different order of the bbox than shapely and geopandas
                minx, miny, maxx, maxy = img.bounds
                self._index.insert(i, (minx, maxx, miny, maxy, mint, maxt))
            return self._index

    def __repr__(self) -> str:
        """String representation."""
        return f"{self.__class__.__name__}({len(self)})"

    @property
    def unary_union(self) -> Polygon | MultiPolygon:
        """(Multi)Polygon representing the union of all image bounds."""
        return unary_union([img.unary_union for img in self])

    @property
    def bounds(self) -> tuple[int, int, int, int]:
        """Total bounds for all Images combined."""
        return get_total_bounds([img.bounds for img in self])

    @property
    def crs(self) -> Any:
        """Common coordinate reference system of the Images."""
        return get_common_crs([img.crs for img in self])


def concat_image_collections(collections: Sequence[ImageCollection]) -> ImageCollection:
    """Union multiple ImageCollections together.

    Same as using the union operator |.
    """
    resolutions = {x.res for x in collections}
    if len(resolutions) > 1:
        raise ValueError(f"resoultion mismatch. {resolutions}")
    images = list(itertools.chain.from_iterable([x.images for x in collections]))
    levels = {x.level for x in collections}
    level = next(iter(levels)) if len(levels) == 1 else None
    first_collection = collections[0]

    out_collection = first_collection.__class__(
        images,
        level=level,
        **first_collection._common_init_kwargs,
        # res=list(resolutions)[0],
    )
    out_collection._all_filepaths = list(
        sorted(
            set(itertools.chain.from_iterable([x._all_filepaths for x in collections]))
        )
    )
    return out_collection


def _get_gradient(band: Band, degrees: bool = False, copy: bool = True) -> Band:
    copied = band.copy() if copy else band
    if len(copied.values.shape) == 3:
        return np.array(
            [_slope_2d(arr, copied.res, degrees=degrees) for arr in copied.values]
        )
    elif len(copied.values.shape) == 2:
        return _slope_2d(copied.values, copied.res, degrees=degrees)
    else:
        raise ValueError("array must be 2 or 3 dimensional")


def to_xarray(
    array: np.ndarray, transform: Affine, crs: Any, name: str | None = None
) -> DataArray:
    """Convert the raster to  an xarray.DataArray."""
    if len(array.shape) == 2:
        height, width = array.shape
        dims = ["y", "x"]
    elif len(array.shape) == 3:
        height, width = array.shape[1:]
        dims = ["band", "y", "x"]
    else:
        raise ValueError(f"Array should be 2 or 3 dimensional. Got shape {array.shape}")

    coords = _generate_spatial_coords(transform, width, height)
    return xr.DataArray(
        array,
        coords=coords,
        dims=dims,
        name=name,
        attrs={"crs": crs},
    )  # .transpose("y", "x")


def _slope_2d(array: np.ndarray, res: int, degrees: int) -> np.ndarray:
    gradient_x, gradient_y = np.gradient(array, res, res)

    gradient = abs(gradient_x) + abs(gradient_y)

    if not degrees:
        return gradient

    radians = np.arctan(gradient)
    degrees = np.degrees(radians)

    assert np.max(degrees) <= 90

    return degrees


def _get_images(
    image_paths: list[str],
    *,
    res: int,
    all_file_paths: list[str],
    df: pd.DataFrame,
    processes: int,
    image_class: Image,
    _mask: GeoDataFrame | GeoSeries | Geometry | tuple[float] | None,
) -> list[Image]:
    with joblib.Parallel(n_jobs=processes, backend="threading") as parallel:
        return parallel(
            joblib.delayed(image_class)(
                path,
                df=df,
                res=res,
                all_file_paths=all_file_paths,
                _mask=_mask,
                processes=processes,
            )
            for path in image_paths
        )


def numpy_to_torch(array: np.ndarray) -> torch.Tensor:
    """Convert numpy array to a pytorch tensor."""
    # fix numpy dtypes which are not supported by pytorch tensors
    if array.dtype == np.uint16:
        array = array.astype(np.int32)
    elif array.dtype == np.uint32:
        array = array.astype(np.int64)

    return torch.tensor(array)


class _RegexError(ValueError):
    pass


class PathlessImageError(ValueError):
    """Exception for when Images, Bands or ImageCollections have no path."""

    def __init__(self, instance: _ImageBase) -> None:
        """Initialise error class."""
        self.instance = instance

    def __str__(self) -> str:
        """String representation."""
        if self.instance._merged:
            what = "that have been merged"
        elif self.isinstance._from_array:
            what = "from arrays"
        elif self.isinstance._from_gdf:
            what = "from GeoDataFrames"

        return (
            f"{self.instance.__class__.__name__} instances {what} "
            "have no 'path' until they are written to file."
        )


def _get_regex_match_from_xml_in_local_dir(
    paths: list[str], regexes: str | tuple[str]
) -> str | dict[str, str]:
    for i, path in enumerate(paths):
        if ".xml" not in path:
            continue
        with open_func(path, "rb") as file:
            filebytes: bytes = file.read()
            try:
                return _extract_regex_match_from_string(
                    filebytes.decode("utf-8"), regexes
                )
            except _RegexError as e:
                if i == len(paths) - 1:
                    raise e


def _extract_regex_match_from_string(
    xml_file: str, regexes: tuple[str]
) -> str | dict[str, str]:
    for regex in regexes:
        if isinstance(regex, dict):
            out = {}
            for key, value in regex.items():
                try:
                    out[key] = re.search(value, xml_file).group(1)
                except (TypeError, AttributeError):
                    continue
            if len(out) != len(regex):
                raise _RegexError()
            return out
        try:
            return re.search(regex, xml_file).group(1)
        except (TypeError, AttributeError):
            continue
    raise _RegexError()


def _fix_path(path: str) -> str:
    return (
        str(path).replace("\\", "/").replace(r"\"", "/").replace("//", "/").rstrip("/")
    )


def _get_regexes_matches_for_df(
    df, match_col: str, patterns: Sequence[re.Pattern], suffix: str = ""
) -> tuple[pd.DataFrame, list[str]]:
    if not len(df):
        return df, []
    assert df.index.is_unique
    matches: list[pd.DataFrame] = []
    for pat in patterns:
        if pat.groups:
            try:
                matches.append(df[match_col].str.extract(pat))
            except ValueError:
                continue
        else:
            match_ = df[match_col].loc[df[match_col].str.match(pat)]
            if len(match_):
                matches.append(match_)

    matches = pd.concat(matches).groupby(level=0, dropna=True).first()

    if isinstance(matches, pd.Series):
        matches = pd.DataFrame({matches.name: matches.values}, index=matches.index)

    match_cols = [f"{col}{suffix}" for col in matches.columns]
    df[match_cols] = matches
    return (
        df.loc[~df[match_cols].isna().all(axis=1)].drop(
            columns=f"{match_col}{suffix}", errors="ignore"
        ),
        match_cols,
    )


def _arr_from_gdf(
    gdf: GeoDataFrame,
    res: int,
    fill: int = 0,
    all_touched: bool = False,
    merge_alg: Callable = MergeAlg.replace,
    default_value: int = 1,
    dtype: Any | None = None,
) -> np.ndarray:
    """Construct Raster from a GeoDataFrame or GeoSeries.

    The GeoDataFrame should have

    Args:
        gdf: The GeoDataFrame to rasterize.
        res: Resolution of the raster in units of the GeoDataFrame's coordinate reference system.
        fill: Fill value for areas outside of input geometries (default is 0).
        all_touched: Whether to consider all pixels touched by geometries,
            not just those whose center is within the polygon (default is False).
        merge_alg: Merge algorithm to use when combining geometries
            (default is 'MergeAlg.replace').
        default_value: Default value to use for the rasterized pixels
            (default is 1).
        dtype: Data type of the output array. If None, it will be
            determined automatically.

    Returns:
        A Raster instance based on the specified GeoDataFrame and parameters.

    Raises:
        TypeError: If 'transform' is provided in kwargs, as this is
        computed based on the GeoDataFrame bounds and resolution.
    """
    if isinstance(gdf, GeoSeries):
        values = gdf.index
        gdf = gdf.to_frame("geometry")
    elif isinstance(gdf, GeoDataFrame):
        if len(gdf.columns) > 2:
            raise ValueError(
                "gdf should have only a geometry column and one numeric column to "
                "use as array values. "
                "Alternatively only a geometry column and a numeric index."
            )
        elif len(gdf.columns) == 1:
            values = gdf.index
        else:
            col: str = next(
                iter([col for col in gdf if col != gdf._geometry_column_name])
            )
            values = gdf[col]

    if isinstance(values, pd.MultiIndex):
        raise ValueError("Index cannot be MultiIndex.")

    shape = _get_shape_from_bounds(gdf.total_bounds, res=res)
    transform = _get_transform_from_bounds(gdf.total_bounds, shape)

    return features.rasterize(
        _gdf_to_geojson_with_col(gdf, values),
        out_shape=shape,
        transform=transform,
        fill=fill,
        all_touched=all_touched,
        merge_alg=merge_alg,
        default_value=default_value,
        dtype=dtype,
    )


def _gdf_to_geojson_with_col(gdf: GeoDataFrame, values: np.ndarray) -> list[dict]:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        return [
            (feature["geometry"], val)
            for val, feature in zip(
                values, loads(gdf.to_json())["features"], strict=False
            )
        ]


def _date_is_within(
    path,
    date_ranges: (
        tuple[str | None, str | None] | tuple[tuple[str | None, str | None], ...] | None
    ),
    image_patterns: Sequence[re.Pattern],
    date_format: str,
) -> bool:
    for pat in image_patterns:
        try:
            date = re.match(pat, Path(path).name).group("date")
            break
        except AttributeError:
            date = None

    if date is None:
        return False

    if date_ranges is None:
        return True

    if all(x is None or isinstance(x, (str, float)) for x in date_ranges):
        date_ranges = (date_ranges,)

    if all(isinstance(x, float) for date_range in date_ranges for x in date_range):
        date = disambiguate_timestamp(date, date_format)
    else:
        date = date[:8]

    for date_range in date_ranges:
        date_min, date_max = date_range

        if isinstance(date_min, float) and isinstance(date_max, float):
            if date[0] >= date_min + 0.0000001 and date[1] <= date_max - 0.0000001:
                return True
            continue

        try:
            date_min = date_min or "00000000"
            date_max = date_max or "99999999"
            assert isinstance(date_min, str)
            assert len(date_min) == 8
            assert isinstance(date_max, str)
            assert len(date_max) == 8
        except AssertionError as err:
            raise TypeError(
                "date_ranges should be a tuple of two 8-charactered strings (start and end date)."
                f"Got {date_range} of type {[type(x) for x in date_range]}"
            ) from err
        if date >= date_min and date <= date_max:
            return True

    return False


def _get_shape_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, res: int
) -> tuple[int, int]:
    resx, resy = (res, res) if isinstance(res, numbers.Number) else res

    minx, miny, maxx, maxy = to_bbox(obj)
    diffx = maxx - minx
    diffy = maxy - miny
    width = int(diffx / resx)
    heigth = int(diffy / resy)
    return heigth, width


def _get_transform_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, shape: tuple[float, ...]
) -> Affine:
    minx, miny, maxx, maxy = to_bbox(obj)
    if len(shape) == 2:
        width, height = shape
    elif len(shape) == 3:
        _, width, height = shape
    else:
        raise ValueError
    return rasterio.transform.from_bounds(minx, miny, maxx, maxy, width, height)


def _get_shape_from_res(
    bounds: tuple[float], res: int, indexes: int | tuple[int]
) -> tuple[int] | None:
    if res is None:
        return None
    if hasattr(res, "__iter__") and len(res) == 2:
        res = res[0]
    diffx = bounds[2] - bounds[0]
    diffy = bounds[3] - bounds[1]
    width = int(diffx / res)
    height = int(diffy / res)
    if not isinstance(indexes, int):
        return len(indexes), width, height
    return width, height


def _array_to_geojson(
    array: np.ndarray, transform: Affine, processes: int
) -> list[tuple]:
    if np.ma.is_masked(array):
        array = array.data
    try:
        return _array_to_geojson_loop(array, transform, processes)

    except ValueError:
        try:
            array = array.astype(np.float32)
            return _array_to_geojson_loop(array, transform, processes)

        except Exception as err:
            raise err.__class__(array.shape, err) from err


def _array_to_geojson_loop(array, transform, processes):
    if processes == 1:
        return [
            (value, shape(geom))
            for geom, value in features.shapes(array, transform=transform, mask=None)
        ]
    else:
        with joblib.Parallel(n_jobs=processes, backend="threading") as parallel:
            return parallel(
                joblib.delayed(_value_geom_pair)(value, geom)
                for geom, value in features.shapes(
                    array, transform=transform, mask=None
                )
            )


def _value_geom_pair(value, geom):
    return (value, shape(geom))


def _intesects(x, other) -> bool:
    return box(*x.bounds).intersects(other)


def _copy_and_add_df_parallel(
    i: tuple[Any, ...], group: pd.DataFrame, self: ImageCollection
) -> tuple[tuple[Any], ImageCollection]:
    copied = self.copy()
    copied.images = [
        img.copy() for img in group.drop_duplicates("_image_idx")["_image_instance"]
    ]
    if "band_id" in group:
        band_ids = set(group["band_id"].values)
        for img in copied.images:
            img._bands = [band for band in img if band.band_id in band_ids]

    # for col in group.columns.difference({"_image_instance", "_image_idx"}):
    #     if not all(
    #         col in dir(band) or col in band.__dict__ for img in copied for band in img
    #     ):
    #         continue
    #     values = set(group[col].values)
    #     for img in copied.images:
    #         img._bands = [band for band in img if getattr(band, col) in values]

    return (i, copied)


def _open_raster(path: str | Path) -> rasterio.io.DatasetReader:
    with opener(path) as file:
        return rasterio.open(file)


def _load_band(band: Band, **kwargs) -> None:
    band.load(**kwargs)


def _merge_by_band(collection: ImageCollection, **kwargs) -> Image:
    print("_merge_by_band", collection.dates)
    return collection.merge_by_band(**kwargs)


def _merge(collection: ImageCollection, **kwargs) -> Band:
    return collection.merge(**kwargs)


def _zonal_one_pair(i: int, poly: Polygon, band: Band, aggfunc, array_func, func_names):
    clipped = band.copy().load(bounds=poly)
    if not np.size(clipped.values):
        return _no_overlap_df(func_names, i, date=band.date)
    return _aggregate(clipped.values, array_func, aggfunc, func_names, band.date, i)


class Sentinel2Config:
    """Holder of Sentinel 2 regexes, band_ids etc."""

    image_regexes: ClassVar[str] = (
        config.SENTINEL2_IMAGE_REGEX,
    )  # config.SENTINEL2_MOSAIC_IMAGE_REGEX,)
    filename_regexes: ClassVar[str] = (
        config.SENTINEL2_FILENAME_REGEX,
        # config.SENTINEL2_MOSAIC_FILENAME_REGEX,
        config.SENTINEL2_CLOUD_FILENAME_REGEX,
    )
    all_bands: ClassVar[list[str]] = list(config.SENTINEL2_BANDS)
    rbg_bands: ClassVar[list[str]] = ["B02", "B03", "B04"]
    ndvi_bands: ClassVar[list[str]] = ["B04", "B08"]
    cloud_band: ClassVar[str] = "SCL"
    cloud_values: ClassVar[tuple[int]] = (3, 8, 9, 10, 11)
    l2a_bands: ClassVar[dict[str, int]] = config.SENTINEL2_L2A_BANDS
    l1c_bands: ClassVar[dict[str, int]] = config.SENTINEL2_L1C_BANDS
    date_format: ClassVar[str] = "%Y%m%d"  # T%H%M%S"


class Sentinel2CloudlessConfig(Sentinel2Config):
    """Holder of regexes, band_ids etc. for Sentinel 2 cloudless mosaic."""

    image_regexes: ClassVar[str] = (config.SENTINEL2_MOSAIC_IMAGE_REGEX,)
    filename_regexes: ClassVar[str] = (config.SENTINEL2_MOSAIC_FILENAME_REGEX,)
    cloud_band: ClassVar[None] = None
    cloud_values: ClassVar[None] = None
    date_format: ClassVar[str] = "%Y%m%d"


class Sentinel2Band(Sentinel2Config, Band):
    """Band with Sentinel2 specific name variables and regexes."""


class Sentinel2Image(Sentinel2Config, Image):
    """Image with Sentinel2 specific name variables and regexes."""

    cloud_cover_regexes: ClassVar[tuple[str]] = config.CLOUD_COVERAGE_REGEXES
    band_class: ClassVar[Sentinel2Band] = Sentinel2Band

    def get_ndvi(
        self,
        red_band: str = Sentinel2Config.ndvi_bands[0],
        nir_band: str = Sentinel2Config.ndvi_bands[1],
    ) -> NDVIBand:
        """Calculate the NDVI for the Image."""
        return super().get_ndvi(red_band=red_band, nir_band=nir_band)


class Sentinel2Collection(Sentinel2Config, ImageCollection):
    """ImageCollection with Sentinel2 specific name variables and regexes."""

    image_class: ClassVar[Sentinel2Image] = Sentinel2Image
    band_class: ClassVar[Sentinel2Band] = Sentinel2Band


class Sentinel2CloudlessBand(Sentinel2CloudlessConfig, Band):
    """Band for cloudless mosaic with Sentinel2 specific name variables and regexes."""


class Sentinel2CloudlessImage(Sentinel2CloudlessConfig, Sentinel2Image):
    """Image for cloudless mosaic with Sentinel2 specific name variables and regexes."""

    # image_regexes: ClassVar[str] = (config.SENTINEL2_MOSAIC_IMAGE_REGEX,)
    # filename_regexes: ClassVar[str] = (config.SENTINEL2_MOSAIC_FILENAME_REGEX,)

    cloud_cover_regexes: ClassVar[None] = None
    band_class: ClassVar[Sentinel2CloudlessBand] = Sentinel2CloudlessBand

    get_ndvi = Sentinel2Image.get_ndvi
    # def get_ndvi(
    #     self,
    #     red_band: str = Sentinel2Config.ndvi_bands[0],
    #     nir_band: str = Sentinel2Config.ndvi_bands[1],
    # ) -> NDVIBand:
    #     """Calculate the NDVI for the Image."""
    #     return super().get_ndvi(red_band=red_band, nir_band=nir_band)


class Sentinel2CloudlessCollection(Sentinel2CloudlessConfig, ImageCollection):
    """ImageCollection with Sentinel2 specific name variables and regexes."""

    # image_regexes: ClassVar[str] = (config.SENTINEL2_MOSAIC_IMAGE_REGEX,)
    # filename_regexes: ClassVar[str] = (config.SENTINEL2_MOSAIC_FILENAME_REGEX,)

    image_class: ClassVar[Sentinel2CloudlessImage] = Sentinel2CloudlessImage
    band_class: ClassVar[Sentinel2Band] = Sentinel2Band
